"""Types for exachanges."""

from __future__ import annotations

from enum import Enum, IntEnum
from typing import TYPE_CHECKING, NotRequired, Optional, TypedDict

from hexbytes import HexBytes

if TYPE_CHECKING:
    from decimal import Decimal

    from eth_typing import ChecksumAddress
    from pandas import Timestamp

TradeResults = HexBytes | dict


class ExchangeType(IntEnum):
    """Exchange type."""

    DEX = 0
    # CEX = 1


class PerpsTradeType(IntEnum):
    """Trade Types."""

    MARKET = 0
    LIMIT = 1
    STOP_MARKET = 2
    STOP_LIMIT = 3
    TRIGGER_TP = 4
    TRIGGER_SL = 5


class PerpsTradeDirection(Enum):
    """Trade direction."""

    SHORT = False
    LONG = True


class PriceData(TypedDict):
    """PriceData from exchange."""

    price: Decimal
    date: Timestamp
    volume: NotRequired[float]


class PriceHistory(TypedDict):
    """Price history from exchange dict."""

    date: list[Timestamp]
    open: list[float]
    high: list[float]
    low: list[float]
    close: list[float]
    volume: NotRequired[float]


class PerpsPosition(TypedDict):
    """Trade position from exchange."""

    pair: str
    id: int
    position_size_stable: Decimal
    collateral_stable: Decimal
    open_price: Decimal
    trade_direction: PerpsTradeDirection
    leverage: Decimal
    liquidation_price: Decimal
    extra: NotRequired[dict]


class PnlDetails(TypedDict):
    """Pnl details from exchange."""

    pnl_usd_before_fees: Decimal
    pnl_percentage_before_fees: Decimal
    funding_fee_usd: Decimal
    position_fee_usd: Decimal
    pnl_usd_after_fees: Decimal
    pnl_percentage_after_fees: Decimal


class OrderData(TypedDict):
    """Order data from exchange."""

    id: str
    pair: str
    trigger_price: Decimal
    size_stable: Decimal
    trade_direction: PerpsTradeDirection
    order_type: PerpsTradeType
    reduce_only: bool
    extra: NotRequired[dict]


class OptionsSortingBy(Enum):
    """Sorting by filters for options."""

    RATE = "rate"
    RESERVED = "reserved"
    AVAILABLE = "available"
    DURATION = "duration"
    PERCENT = "percent"
    AMOUNT = "AMOUNT"

    def __str__(self) -> str:
        """Override str to return value."""
        return self.value


class OptionsSortingDestination(Enum):
    """Sorting destination filter for options based on sortingBy."""

    ASC = "ASC"
    DESC = "DESC"

    def __str__(self) -> str:
        """Override str to return value."""
        return self.value


class OptionsOrderType(Enum):
    """Order type for options."""

    MY_ORDER = "my_order"
    ALL_ORDER = "all_order"

    def __str__(self) -> str:
        """Override str to return value."""
        return self.value


class OptionsDirection(Enum):
    """Options direction."""

    UP = "up"
    DOWN = "down"

    def __str__(self) -> str:
        """Override str to return value."""
        return self.value


class OptionsDuration(IntEnum):
    """Duration argument for Options."""

    M15 = 15 * 60
    M30 = 30 * 60
    H1 = 60 * 60
    H2 = 2 * 60 * 60
    H4 = 4 * 60 * 60
    H8 = 8 * 60 * 60
    H24 = 24 * 60 * 60


class OptionsPercent(IntEnum):
    """Percent for options."""

    UP = 1000000100000000000
    UP_010 = 1001000000000000000
    UP_025 = 1002500000000000000
    UP_050 = 1005000000000000000
    UP_1 = 1010000000000000000
    UP_3 = 1030000000000000000
    UP_5 = 1050000000000000000
    DOWN = 999999900000000000
    DOWN_010 = 999000000000000000
    DOWN_025 = 997500000000000000
    DOWN_050 = 995000000000000000
    DOWN_1 = 990000000000000000
    DOWN_3 = 970000000000000000
    DOWN_5 = 950000000000000000


class OptionsOrdersParams(TypedDict, total=False):
    """Params for options orders query."""

    oracle_ids: Optional[list[int]]
    sorting_by: Optional[OptionsSortingBy]
    sorting_destination: Optional[OptionsSortingDestination]
    closed: Optional[bool]
    account: Optional[ChecksumAddress]
    order_type: Optional[OptionsOrderType]
    duration: Optional[OptionsDuration]
    percent: Optional[str]
    skip: Optional[int]
    limit: Optional[int]


class OptionsRisk(Enum):
    """Options risks strategy."""

    PRIO_RETURN = 0
    PRIO_SAFETY = 1


class OptionsStrategy(TypedDict):
    """Strategy params for options buy."""

    rate_min: float
    available_min: float
    percent_min: OptionsPercent
    percent_max: OptionsPercent
    duration_min: OptionsDuration
    duration_max: OptionsDuration
    risk: OptionsRisk
    direction: OptionsDirection


class OptionsBuyParams(TypedDict):
    """Buy params for options."""

    orders: dict[int, int]
    price_id: str


class NewAccountInfo(TypedDict):
    """New account info."""

    referral_link: NotRequired[Optional[str]]
    secrets: list[str]



"""Class to manage app config."""

from typing import Any

import keyring
import orjson
from peewee import ModelBase

from plutus_terminal.core.db.models import (
    DATABASE,
    DATABASE_PATH,
    GUISettings,
    KeyringAccount,
    OptionsConfig,
    TradeConfig,
    UserFilter,
    Web3RPC,
    create_database,
)
from plutus_terminal.core.types_ import ExchangeType, OptionsDuration, OptionsRisk


class AppConfig:
    """Manage app config."""

    SERVICE_NAME = "plutus_terminal"

    def __init__(self) -> None:
        """Initialize shared variables."""
        self._current_keyring_account: KeyringAccount
        self._gui_settings_cache: dict[str, Any] = {}

        self._leverage = 0
        self._stop_loss = 0.0
        self._take_profit = 0.0
        self._trade_value_lowest = 0
        self._trade_value_low = 0
        self._trade_value_medium = 0
        self._trade_value_high = 0

        self._options_rate_min = 0.0
        self._options_amount = 0.0
        self._options_avaialble_min = 0.0
        self._options_percent_min = ""
        self._options_percent_max = ""
        self._options_duration_min = 0
        self._options_duration_max = 0
        self._options_risk = 0

        self._validate_database()

    @property
    def current_keyring_account(self) -> KeyringAccount:
        """Returns current account set in the terminal."""
        return self._current_keyring_account

    @current_keyring_account.setter
    def current_keyring_account(self, new_account: KeyringAccount) -> None:
        """Set new current account."""
        if new_account not in self.get_all_keyring_accounts():
            msg = f"Invalid account: {new_account}"
            raise ValueError(msg)
        with DATABASE.atomic():
            self.set_gui_settings("current_account_id", new_account.id)  # type: ignore

        self._current_keyring_account = new_account
        self.load_config()

    @property
    def leverage(self) -> int:
        """Returns leverage to be used in the terminal.

        This value is associated with the current account.
        """
        return self._leverage

    @leverage.setter
    def leverage(self, new_value: int) -> None:
        """Set new leverage value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"leverage": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._leverage = new_value

    @property
    def stop_loss(self) -> float:
        """Returns stop_loss to be used in the terminal.

        This value is associated with the current account.
        """
        return self._stop_loss

    @stop_loss.setter
    def stop_loss(self, new_value: float) -> None:
        """Set new stop_loss value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"stop_loss": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._stop_loss = new_value

    @property
    def take_profit(self) -> float:
        """Returns take_profit to be used in the terminal.

        This value is associated with the current account.
        """
        return self._take_profit

    @take_profit.setter
    def take_profit(self, new_value: float) -> None:
        """Set new take_profit value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"take_profit": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._take_profit = new_value

    @property
    def trade_value_lowest(self) -> int:
        """Returns trade_value_lowest to be used in the terminal.

        This value is associated with the current account.
        """
        return self._trade_value_lowest

    @trade_value_lowest.setter
    def trade_value_lowest(self, new_value: int) -> None:
        """Set new trade_value_lowest value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"trade_value_lowest": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._trade_value_lowest = new_value

    @property
    def trade_value_low(self) -> int:
        """Returns trade_value_lowe to be used in the terminal.

        This value is associated with the current account.
        """
        return self._trade_value_low

    @trade_value_low.setter
    def trade_value_low(self, new_value: int) -> None:
        """Set new trade_value_low value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"trade_value_low": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._trade_value_low = new_value

    @property
    def trade_value_medium(self) -> int:
        """Returns trade_value_medium to be used in the terminal.

        This value is associated with the current account.
        """
        return self._trade_value_medium

    @trade_value_medium.setter
    def trade_value_medium(self, new_value: int) -> None:
        """Set new trade_value_medium value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"trade_value_medium": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._trade_value_medium = new_value

    @property
    def trade_value_high(self) -> int:
        """Returns trade_value_high to be used in the terminal.

        This value is associated with the current account.
        """
        return self._trade_value_high

    @trade_value_high.setter
    def trade_value_high(self, new_value: int) -> None:
        """Set new trade_value_high value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                TradeConfig,
                {"trade_value_high": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._trade_value_high = new_value

    @property
    def options_amount(self) -> float:
        """Returns amount to be used in the terminal.

        This value is associated with the current account.
        """
        return self._options_amount

    @options_amount.setter
    def options_amount(self, new_value: float) -> None:
        """Set new amount value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"amount": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_amount = new_value

    @property
    def options_rate_min(self) -> float:
        """Returns rate_min to be used in the terminal.

        This value is associated with the current account.
        """
        return self._options_rate_min

    @options_rate_min.setter
    def options_rate_min(self, new_value: float) -> None:
        """Set new rate_min value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"rate_min": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_rate_min = new_value

    @property
    def options_available_min(self) -> float:
        """Returns available_min to be used in the terminal.

        This value is associated with the current account.
        """
        return self._options_avaialble_min

    @options_available_min.setter
    def options_available_min(self, new_value: float) -> None:
        """Set new available_min value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"available_min": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_avaialble_min = new_value

    @property
    def options_percent_min(self) -> str:
        """Returns percent_min to be used in the terminal.

        This value is associated with the current account.
        """
        return self._options_percent_min

    @options_percent_min.setter
    def options_percent_min(self, new_value: str) -> None:
        """Set new percent_min value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"percent_min": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_percent_min = new_value

    @property
    def options_percent_max(self) -> str:
        """Returns percent_max to be used in the terminal.

        This value is associated with the current account.
        """
        return self._options_percent_max

    @options_percent_max.setter
    def options_percent_max(self, new_value: str) -> None:
        """Set new percent_max value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"percent_max": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_percent_max = new_value

    @property
    def options_duration_min(self) -> OptionsDuration:
        """Returns duration_min to be used in the terminal.

        This value is associated with the current account.
        """
        return OptionsDuration(self._options_duration_min)

    @options_duration_min.setter
    def options_duration_min(self, new_value: int) -> None:
        """Set new duration_min value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"duration_min": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_duration_min = new_value

    @property
    def options_duration_max(self) -> OptionsDuration:
        """Returns duration_max to be used in the terminal.

        This value is associated with the current account.
        """
        return OptionsDuration(self._options_duration_max)

    @options_duration_max.setter
    def options_duration_max(self, new_value: int) -> None:
        """Set new duration_max value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"duration_max": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_duration_max = new_value

    @property
    def options_risk(self) -> OptionsRisk:
        """Returns risk to be used in the terminal.

        This value is associated with the current account.
        """
        return OptionsRisk(self._options_risk)

    @options_risk.setter
    def options_risk(self, new_value: int) -> None:
        """Set new risk value for current_keyring_account."""
        with DATABASE.atomic():
            self._generic_update(
                OptionsConfig,
                {"risk": new_value},
                self.current_keyring_account.id,  # type: ignore
            )
        self._options_risk = new_value

    def _generic_update(
        self,
        model: ModelBase,
        attribute_dict: dict[str, Any],
        model_id: int,
    ) -> None:
        """Update Model attribute with given vallue.

        Args:
            model (Peewee.Model): Model to update value.
            attribute_dict (dict[str, Any]): Dict with attribute name and value.
            model_id (int): Id of the model to update.
        """
        query = model.update(**attribute_dict).where(model.id == model_id)  # type: ignore
        query.execute()

    def load_config(self) -> None:
        """Load config values from database."""
        account_id = GUISettings.get(GUISettings.key == "current_account_id").value

        keyring_account: KeyringAccount = KeyringAccount.get_by_id(account_id)
        self._current_keyring_account = keyring_account

        trade_config = TradeConfig.get(TradeConfig.account == keyring_account.id)  # type: ignore

        self._leverage = trade_config.leverage
        self._stop_loss = trade_config.stop_loss
        self._take_profit = trade_config.take_profit
        self._trade_value_lowest = trade_config.trade_value_lowest
        self._trade_value_low = trade_config.trade_value_low
        self._trade_value_medium = trade_config.trade_value_medium
        self._trade_value_high = trade_config.trade_value_high

        options_config = OptionsConfig.get(OptionsConfig.account == keyring_account.id)  # type: ignore
        self._options_amount = options_config.amount
        self._options_rate_min = options_config.rate_min
        self._options_avaialble_min = options_config.available_min
        self._options_percent_min = options_config.percent_min
        self._options_percent_max = options_config.percent_max
        self._options_duration_min = options_config.duration_min
        self._options_duration_max = options_config.duration_max
        self._options_risk = options_config.risk

        # create default web3
        self.create_default_rpcs()

    def get_gui_settings(self, key: str) -> Any:  # noqa: ANN401
        """Get GUI settings value for key."""
        cached_value = self._gui_settings_cache.get(key, None)
        if cached_value is not None:
            return cached_value

        value = GUISettings.get(GUISettings.key == key).value
        value = orjson.loads(value)
        self._gui_settings_cache[key] = value
        return value

    def set_gui_settings(self, key: str, value: Any) -> None:  # noqa: ANN401
        """Set GUI settings value for key."""
        self._gui_settings_cache[key] = value
        value = orjson.dumps(value)
        GUISettings.update(value=value).where(GUISettings.key == key).execute()

    def _validate_database(self) -> None:
        """Validate database existence and tables."""
        if not DATABASE_PATH.exists():
            create_database()

    @staticmethod
    def get_all_keyring_accounts() -> list[KeyringAccount]:
        """Get all keyring accounts."""
        with DATABASE.atomic():
            return KeyringAccount.select()

    @staticmethod
    def get_all_user_filters() -> list[UserFilter]:
        """Get all user filters from database."""
        with DATABASE.atomic():
            return UserFilter.select()

    @staticmethod
    def write_model_to_db(model: Any) -> None:  # noqa: ANN401
        """Write model to database."""
        with DATABASE.atomic():
            model.save()

    @staticmethod
    def delete_user_filter(user_filter_id: int) -> None:
        """Delete user_filter from database."""
        with DATABASE.atomic():
            UserFilter.delete().where(UserFilter.id == user_filter_id).execute()  # type: ignore

    @staticmethod
    def create_account(
        username: str,
        exchange_type: ExchangeType,
        exchange_name: str,
    ) -> KeyringAccount:
        """Create new account."""
        with DATABASE:
            keyring_account = KeyringAccount.create(
                username=username,
                exchange_type=exchange_type,
                exchange_name=exchange_name,
            )
            TradeConfig.create(account=keyring_account)
            OptionsConfig.create(account=keyring_account)
        return keyring_account

    @staticmethod
    def delete_account(account_id: int) -> None:
        """Delete account."""
        with DATABASE.atomic():
            keyring_account = KeyringAccount.get_by_id(account_id)
            keyring.delete_password("plutus-terminal", str(keyring_account.username))
            KeyringAccount.delete().where(KeyringAccount.id == account_id).execute()  # type: ignore
            TradeConfig.delete().where(TradeConfig.account == account_id).execute()
            OptionsConfig.delete().where(OptionsConfig.account == account_id).execute()

    @staticmethod
    def create_default_rpcs() -> None:
        """Create default values for Web3 RPC."""
        with DATABASE.atomic():
            # Arbitrum One
            Web3RPC.get_or_create(
                chain_name="Arbitrum One Fetcher",
                defaults={
                    "rpc_urls": orjson.dumps(
                        [
                            "https://arbitrum-one-rpc.publicnode.com",
                            "https://arbitrum.blockpi.network/v1/rpc/public",
                            "https://rpc.ankr.com/arbitrum",
                            "https://arbitrum-one.public.blastapi.io/",
                            "https://arbitrum.llamarpc.com",
                        ],
                    ),
                },
            )

            Web3RPC.get_or_create(
                chain_name="Arbitrum One Trader",
                defaults={
                    "rpc_urls": orjson.dumps(
                        [
                            "https://arb1.arbitrum.io/rpc",
                        ],
                    ),
                },
            )

    @staticmethod
    def get_web3_rpc_by_name(chain_name: str) -> Web3RPC:
        """Get Web3 RPC by name."""
        with DATABASE.atomic():
            return Web3RPC.get(Web3RPC.chain_name == chain_name)

    @staticmethod
    def get_all_web3_rpc() -> list[Web3RPC]:
        """Get all Web3 RPC."""
        with DATABASE.atomic():
            return Web3RPC.select()


CONFIG = AppConfig()



"""Protocols and base class for news fetching."""

from __future__ import annotations

from typing import Protocol

from PySide6.QtCore import QObject, Signal

from plutus_terminal.core.types_ import NewsData


class NewsMessageBus(QObject):
    """Message bus for news message."""

    raw_news_signal = Signal(NewsData)
    news_signal = Signal(NewsData)


class NewsFetcher(Protocol):
    """News fetcher protocol."""

    async def subscribe_to_wss(self, message_bus: NewsMessageBus) -> None:
        """Subscribe to news wss and emit news signal on new entry.

        Args:
            message_bus (plutus_terminal.ui.thread.NewsMessageBus): Message bus
                to emit news messages
        """
        ...

    async def fetch_old_news(self, limit: int) -> list[NewsData]:
        """Fetch old news from API.

        Args:
            limit (int): Amount of news to fetch.

        Returns:
            list[NewsData]: List of old news. This list is expected to be ordered.
            from latest to oldest.
        """
        ...

    async def login(self) -> None:
        """Login to news source."""
        ...

    async def stop_async(self) -> None:
        """Stop infinite loops and close connections."""
        ...



"""Widget to control web3 configs."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import orjson
from PySide6 import QtCore, QtWidgets

from plutus_terminal.core.config import CONFIG
from plutus_terminal.ui.widgets.toast import Toast, ToastType
from plutus_terminal.ui.widgets.top_bar_widget import TopBar

if TYPE_CHECKING:
    from plutus_terminal.core.db.models import Web3RPC


class Web3Config(QtWidgets.QWidget):
    """Widget to control web3 configs."""

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)

        self._main_layout = QtWidgets.QVBoxLayout()

        self._web3_rpc_bar = TopBar("Web3 RPCs")
        self._rpcs_layout = QtWidgets.QVBoxLayout()
        self._reset_rpcs_btn = QtWidgets.QPushButton("Reset RPCs")
        self._save_rpcs_btn = QtWidgets.QPushButton("Save RPCs")

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self._reset_rpcs_btn.setMinimumSize(80, 30)
        self._reset_rpcs_btn.clicked.connect(self._reset_rpcs)

        self._save_rpcs_btn.setProperty("class", "LONG")
        self._save_rpcs_btn.setMinimumSize(80, 30)
        self._save_rpcs_btn.clicked.connect(self._save_rpcs)

        web3_rcps = CONFIG.get_all_web3_rpc()

        for web3_rpc in web3_rcps:
            rpc_config = RPCConfig(web3_rpc=web3_rpc)
            self._rpcs_layout.addWidget(rpc_config)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addWidget(self._web3_rpc_bar)
        self._main_layout.addLayout(self._rpcs_layout)
        rpcs_layout = QtWidgets.QHBoxLayout()
        rpcs_layout.addStretch()
        rpcs_layout.addWidget(self._reset_rpcs_btn)
        rpcs_layout.addWidget(self._save_rpcs_btn)
        self._main_layout.addLayout(rpcs_layout)
        self._main_layout.addStretch()

        self.setLayout(self._main_layout)

    def _save_rpcs(self) -> None:
        """Save RPCs."""
        for index in range(self._rpcs_layout.count()):
            rpc_config = self._rpcs_layout.itemAt(index).widget()
            if isinstance(rpc_config, RPCConfig):
                rpc_config.write_to_db()

        Toast.show_message("RPCs updated", type_=ToastType.SUCCESS)

    def _reset_rpcs(self) -> None:
        """Reset RPCs."""
        for index in range(self._rpcs_layout.count()):
            rpc_config = self._rpcs_layout.itemAt(index).widget()
            if isinstance(rpc_config, RPCConfig):
                rpc_config.reset_to_current()


class RPCConfig(QtWidgets.QFrame):
    """Widget to control RPC configs."""

    def __init__(
        self,
        web3_rpc: Web3RPC,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)
        self._web3_rpc = web3_rpc

        self._main_layout = QtWidgets.QGridLayout()
        self._label = QtWidgets.QLabel()
        self._model = QtCore.QStringListModel()
        self._list = QtWidgets.QListView()
        self._input_field = QtWidgets.QLineEdit()
        self._add_button = QtWidgets.QPushButton("Add")
        self._remove_button = QtWidgets.QPushButton("Remove")

        self._setup_widgets()
        self._setup_layout()

        self.reset_to_current()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.setObjectName("filter")
        self._label.setObjectName("title")

        self._list.setModel(self._model)
        self._list.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self._input_field.setPlaceholderText("Enter new RPC url...")

        self._add_button.setMinimumHeight(30)
        self._add_button.setProperty("class", "LONG")
        self._add_button.clicked.connect(self._add_rpc)
        self._add_button.setToolTip("Add new RPC from input")
        self._remove_button.setMinimumHeight(30)
        self._remove_button.setProperty("class", "SHORT")
        self._remove_button.clicked.connect(self._remove_rpc)
        self._remove_button.setToolTip("Remove selected RPC")

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addWidget(
            self._label,
            0,
            0,
            1,
            2,
            alignment=QtCore.Qt.AlignmentFlag.AlignCenter,
        )
        self._main_layout.addWidget(self._list, 1, 0, 1, 2)
        self._main_layout.addWidget(self._input_field, 2, 0, 1, 2)
        self._main_layout.addWidget(self._add_button, 3, 0)
        self._main_layout.addWidget(self._remove_button, 3, 1)

        self.setLayout(self._main_layout)

    def reset_to_current(self) -> None:
        """Reset to current."""
        self._label.setText(str(self._web3_rpc.chain_name))
        self._model.setStringList(orjson.loads(str(self._web3_rpc.rpc_urls)))

    def _add_rpc(self) -> None:
        """Add RPC from input to list."""
        current_list = self._model.stringList()
        new_item = self._input_field.text()
        if new_item not in current_list:
            current_list.append(new_item)
            self._model.setStringList(current_list)
        self._input_field.clear()

    def _remove_rpc(self) -> None:
        """Remove selected RPC."""
        selected_indexes = self._list.selectedIndexes()
        if not selected_indexes:
            return
        current_list = self._model.stringList()
        indexes = [sel.row() for sel in selected_indexes]
        for index in sorted(indexes, reverse=True):
            current_list.pop(index)
        self._model.setStringList(current_list)

    def write_to_db(self) -> None:
        """Write web3_rpc to database."""
        current_list = self._model.stringList()
        self._web3_rpc.rpc_urls = orjson.dumps(current_list)  # type: ignore
        CONFIG.write_model_to_db(self._web3_rpc)



"""Base trader for CCXT."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import ccxt.pro

from plutus_terminal.core.exceptions import TransactionFailedError
from plutus_terminal.core.exchange.base import ExchangeTrader
from plutus_terminal.core.types_ import PerpsTradeDirection

if TYPE_CHECKING:
    import plutus_terminal.core.exchange.ccxt.utils as ccxt_utils
    from plutus_terminal.core.exchange.types import TradeResults

LOGGER = logging.getLogger(__name__)


class CCXTTrader(ExchangeTrader):
    """Base trader for CCXT."""

    def __init__(self, exchange_instance: ccxt.pro.Exchange) -> None:
        """Initialize shared attributes."""
        self._cctx = exchange_instance

    async def create_order(
        self,
        trade_arguments: ccxt_utils.TradingArgs,
    ) -> TradeResults:
        """Create new order.

        Args:
            trade_arguments (ccxt.utils.TradingArgs): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        LOGGER.info("Opening new position: %s", trade_arguments)
        side = "buy" if trade_arguments["side"] == PerpsTradeDirection.LONG else "sell"
        trade_type = trade_arguments["trade_type"].name.lower()
        params = {}

        # Add stop loss and take profit if available
        try:
            if trade_arguments["take_profit"] != trade_arguments["price"]:
                params["takeProfit"] = {
                    "type": "market",
                    "triggerPrice": float(trade_arguments["take_profit"]),
                }
        except KeyError:
            pass
        try:
            if trade_arguments["stop_loss"] != trade_arguments["price"]:
                params["stopLoss"] = {
                    "type": "market",
                    "triggerPrice": float(trade_arguments["stop_loss"]),
                }
        except KeyError:
            pass

        try:
            result = await self._cctx.create_order(
                symbol=trade_arguments["pair"],
                type=trade_type,  # type: ignore
                side=side,
                amount=float(trade_arguments["amount"]),
                price=float(trade_arguments["price"]),
                params=params,
            )
        except (
            ccxt.InsufficientFunds,
            ccxt.NetworkError,
        ) as error:
            raise TransactionFailedError from error

        LOGGER.info("Opened position. Result: %s", result)
        return result

    async def close_position(self, trade_arguments: dict) -> TradeResults:
        """Close the give position.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """
        try:
            result = await self._cctx.close_position(
                symbol=trade_arguments["pair"],
            )
        except (
            ccxt.InsufficientFunds,
            ccxt.NetworkError,
        ) as error:
            raise TransactionFailedError from error

        LOGGER.info("Closed position. Result: %s", result)
        return result

    async def cancel_order(self, trade_arguments: dict) -> TradeResults:
        """Cancel the given order.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """

    async def edit_order(
        self,
        trade_arguments: dict,
    ) -> TradeResults:
        """Update the given order.

        Args:
            trade_arguments (dict): Arguments necessary for the trade.

        Raises:
            TransactionFailed: If the transaction fails.

        Returns:
            TradeResults: Result of the trade.
        """



"""Filter manager."""

from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING

from httpx import Client
import orjson

from plutus_terminal.core.config import AppConfig
from plutus_terminal.core.news.filter._actions import FILTER_ACTIONS_MAP
from plutus_terminal.core.news.filter._filters import (
    DataMatchingFilter,
    KeywordMatchingFilter,
)
from plutus_terminal.core.news.filter.types import ActionType, FilterType

if TYPE_CHECKING:
    from plutus_terminal.core.types_ import NewsData


class FilterManager:
    """Manage all filters."""

    def __init__(self) -> None:
        """Initialize."""
        self._keyword_filters = KeywordMatchingFilter()
        self._data_matching_filters = DataMatchingFilter()

        self._filter_type_map = OrderedDict(
            {
                FilterType.KEYWORD_MATCHING: self._keyword_filters,
                FilterType.DATA_MATCHING: self._data_matching_filters,
            },
        )
        self._all_user_filters = AppConfig.get_all_user_filters()

        self._create_internal_filters()
        self._create_user_filters()

    def _fetch_all_token_data(self) -> dict:
        """Fetch all token data.

        Token map for filtering provided by PhoenixNews API.
        """
        with Client() as client:
            response = client.get("https://api.phoenixnews.io/getAllTokens")
        response.raise_for_status()
        return response.json()

    def _create_internal_filters(self) -> None:
        """Create internal filters."""
        # Coin association with token data from PhoenixNews API
        token_data = self._fetch_all_token_data()
        for item in token_data:
            for word in item["baseCurrencyName"]:
                symbol = item["baseSymbol"].replace("\\", "")
                symbol = symbol.replace("$", "")
                self._keyword_filters.add_to_queue(
                    {"keyword": word.lower()},
                    FILTER_ACTIONS_MAP[ActionType.COIN_ASSOCIATION],
                    {"coin": symbol, "color": (255, 140, 0)},
                )

    def _create_user_filters(self) -> None:
        """Create user filters."""
        for user_filter in self._all_user_filters:
            filter_type = FilterType(user_filter.filter_type)  # type: ignore
            match_pattern = orjson.loads(user_filter.match_pattern)  # type: ignore
            action_type = ActionType(user_filter.action_type)  # type: ignore
            action_args = orjson.loads(user_filter.action_args)  # type: ignore

            self._filter_type_map[filter_type].add_to_queue(
                match_pattern,
                FILTER_ACTIONS_MAP[action_type],
                action_args,
            )

    def filter(self, news_data: NewsData) -> NewsData:
        """Run all created filters on the news data provided.

        Args:
            news_data (NewsData): News data to be filtered.

        Returns:
            NewsData: Filtered news data.
        """
        for filter_type in self._filter_type_map.values():
            news_data = filter_type.execute(news_data)
            # Break ealier if this news is being ignored.
            if news_data["ignored"]:
                break
        return news_data



"""New account dialog."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import keyring
import orjson
from PySide6 import QtWidgets
from PySide6.QtCore import QRegularExpression, Qt
from PySide6.QtGui import QPixmap, QRegularExpressionValidator

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exchange.valid_exchanges import VALID_EXCHANGES
from plutus_terminal.core.types_ import ExchangeType
from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    from plutus_terminal.core.db.models import KeyringAccount


class NewAccountDialog(QtWidgets.QDialog):
    """Dialog for creating new account."""

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)

        self.main_layout = QtWidgets.QVBoxLayout(self)
        self.new_account: Optional[KeyringAccount] = None

        self._add_acc_icon = QtWidgets.QLabel()
        self._info_group = QtWidgets.QGroupBox("Account Info:")
        self._info_layout = QtWidgets.QGridLayout()
        self._type_label = QtWidgets.QLabel("Exchange Type")
        self._type_combo_box = QtWidgets.QComboBox()
        self._exchange_label = QtWidgets.QLabel("Exchange")
        self._exchange_combo_box = QtWidgets.QComboBox()
        self._account_label = QtWidgets.QLabel("Account Name")
        self._account_line_edit = QtWidgets.QLineEdit()

        self._secrets_group = QtWidgets.QGroupBox("Secrets:")
        self._secrets_layout = QtWidgets.QGridLayout()
        self._secrets_labels: list[QtWidgets.QLabel] = []
        self._secrets_line_edits: list[QtWidgets.QLineEdit] = []

        self._log_label = QtWidgets.QLabel()
        self._create_btn = QtWidgets.QPushButton("Create New Account")

        self._setup_widgets()
        self._setup_connections()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Configure widgets."""
        self.setWindowTitle("Create New Account")
        self.setWindowIcon(QPixmap(":/icons/plutus_icon"))
        self.setMinimumSize(500, 500)

        self._add_acc_icon.setPixmap(QPixmap(":/icons/new_acc"))
        self._add_acc_icon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._add_acc_icon.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )

        self._info_group.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        self._type_combo_box.addItems(
            [exchange_type.name for exchange_type in ExchangeType],
        )
        self._fill_exchange_options()

        self._secrets_group.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Minimum,
        )

        self._account_line_edit.setPlaceholderText("Enter account name...")
        # Only allow letters, numbers, and underscores up to 24 characters
        regex = QRegularExpression("^[A-Za-z0-9_]{1,24}$")
        self._account_line_edit.setValidator(QRegularExpressionValidator(regex, self))

        self._log_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self._log_label.setProperty("class", "error")

        self._create_btn.setMinimumSize(150, 40)
        self._create_btn.setProperty("class", "LONG")

    def _setup_connections(self) -> None:
        """Configure connections."""
        self._type_combo_box.currentTextChanged.connect(self._fill_exchange_options)
        self._exchange_combo_box.currentTextChanged.connect(self._fill_secrets_group)
        self._create_btn.clicked.connect(self._create_new_account)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self.main_layout.addWidget(self._add_acc_icon)
        self._info_layout.addWidget(self._type_label, 0, 0)
        self._info_layout.addWidget(self._type_combo_box, 0, 1)
        self._info_layout.addWidget(self._exchange_label, 1, 0)
        self._info_layout.addWidget(self._exchange_combo_box, 1, 1)
        self._info_layout.addWidget(self._account_label, 2, 0)
        self._info_layout.addWidget(self._account_line_edit, 2, 1)
        self._info_group.setLayout(self._info_layout)
        self.main_layout.addWidget(self._info_group)

        self._secrets_group.setLayout(self._secrets_layout)
        self.main_layout.addWidget(self._secrets_group)
        self.main_layout.addWidget(self._log_label)
        self.main_layout.addWidget(
            self._create_btn,
            alignment=Qt.AlignmentFlag.AlignRight,
        )

    def _fill_exchange_options(self) -> None:
        """Fill exchange combo box."""
        self._exchange_combo_box.blockSignals(True)

        self._exchange_combo_box.clear()
        selected_type = self._type_combo_box.currentText()

        for exchange_name, exchange in VALID_EXCHANGES.items():
            if exchange.exchange_type() == ExchangeType[selected_type]:
                icon = QPixmap(f":/exchanges/{exchange_name}")
                self._exchange_combo_box.addItem(icon, exchange.name(), userData=exchange_name)

        self._fill_secrets_group()

        self._exchange_combo_box.blockSignals(False)

    def _fill_secrets_group(self) -> None:
        """Fill secrets group."""
        # Clear _secrets_layout widgets
        while self._secrets_layout.count():
            old_widget = self._secrets_layout.takeAt(
                self._secrets_layout.count() - 1,
            ).widget()
            old_widget.deleteLater()

        self._secrets_labels.clear()
        self._secrets_line_edits.clear()

        selected_exchange = self._exchange_combo_box.currentData()
        exchange = VALID_EXCHANGES[selected_exchange]
        new_account_info = exchange.new_account_info()

        for secret in new_account_info["secrets"]:
            label = QtWidgets.QLabel(secret)
            line_edit = QtWidgets.QLineEdit()
            self._secrets_layout.addWidget(label, len(self._secrets_labels), 0)
            self._secrets_layout.addWidget(line_edit, len(self._secrets_labels), 1)
            self._secrets_labels.append(label)
            self._secrets_line_edits.append(line_edit)

    def _create_new_account(self) -> None:
        """Create new account on database."""
        # Validate if all fields are filled
        toast_id = Toast.show_message(
            "Creating account...",
            type_=ToastType.MESSAGE,
        )
        if not self._account_line_edit.text():
            self._log_label.setText("Account name cannot be empty!")
            Toast.update_message(
                toast_id,
                "Account name cannot be empty!",
                ToastType.ERROR,
            )
            return
        for line_edit in self._secrets_line_edits:
            if not line_edit.text():
                self._log_label.setText("Secrets cannot be empty!")
                Toast.update_message(
                    toast_id,
                    "Secrets cannot be empty!",
                    ToastType.ERROR,
                )
                return

        account_name = self._account_line_edit.text()
        secrets = [secret.text() for secret in self._secrets_line_edits]
        exchange_name = self._exchange_combo_box.currentData()

        # Validate secrets
        exchange = VALID_EXCHANGES[exchange_name]
        valid, error = exchange.validate_secrets(secrets)
        if not valid:
            self._log_label.setText(error)
            Toast.update_message(toast_id, error, ToastType.ERROR)
            return

        self.new_account = CONFIG.create_account(
            username=account_name,
            exchange_type=ExchangeType[self._type_combo_box.currentText()],
            exchange_name=exchange_name,
        )
        CONFIG.current_keyring_account = self.new_account

        # Save secrets to keyring
        keyring.set_password(
            "plutus-terminal",
            str(account_name),
            orjson.dumps(secrets).decode("utf-8"),
        )

        self._log_label.setText("Account created successfully!")
        Toast.update_message(
            toast_id,
            "Account created successfully!",
            ToastType.SUCCESS,
        )
        super().accept()



"""Spin Box that holds Decimal values."""

from decimal import Decimal
from typing import Any, Optional

from PySide6 import QtCore, QtGui, QtWidgets


class DecimalSpinBox(QtWidgets.QDoubleSpinBox):
    """Double spin box."""

    decimalValueChanged = QtCore.Signal(Decimal)  # noqa: N815

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.
        """
        self._current_decimal: Decimal = Decimal(0)
        self._maximum_decimal: Decimal = Decimal(100_000_000_000)
        self._minimum_decimal: Decimal = Decimal(-100_000_000_000)
        super().__init__(*args, **kwargs)
        self.setGroupSeparatorShown(True)
        self.setMaximum(self._maximum_decimal)
        self.setMinimum(self._minimum_decimal)
        self.setDecimals(8)
        self.valueChanged.connect(self._on_value_changed)

    def value(self) -> Decimal:  # type: ignore
        """Get current value.

        Returns:
            Decimal: Current value.
        """
        return self._current_decimal

    def setValue(self, value: Decimal) -> None:  # type: ignore
        """Set value.

        Args:
            value (Decimal): Value to set.
        """
        self._current_decimal = value
        self.decimalValueChanged.emit(value)
        self.blockSignals(True)
        super().setValue(float(value))
        self.blockSignals(False)

    def setMaximum(self, max_value: Decimal) -> None:  # type: ignore
        """Set maximum.

        Args:
            max_value (Decimal): Maximum value.
        """
        self._maximum_decimal = max_value
        return super().setMaximum(float(max_value))

    def maximum(self) -> Decimal:  # type: ignore
        """Get maximum.

        Returns:
            Decimal: Maximum value.
        """
        return self._maximum_decimal

    def setMinimum(self, min_value: Decimal) -> None:  # type: ignore
        """Set minimum.

        Args:
            min_value (Decimal): Minimum value.
        """
        self._minimum_decimal = min_value
        return super().setMinimum(float(min_value))

    def minimum(self) -> Decimal:  # type: ignore
        """Get minimum.

        Returns:
            Decimal: Minimum value.
        """
        return self._minimum_decimal

    def setRange(self, min_value: Decimal, max_value: Decimal) -> None:  # type: ignore
        """Set range.

        Args:
            min_value (Decimal): Minimum value.
            max_value (Decimal): Maximum value.
        """
        self._minimum_decimal = min_value
        self._maximum_decimal = max_value
        return super().setRange(float(min_value), float(max_value))

    def _on_value_changed(self, value: float) -> None:
        """On value changed."""
        self._current_decimal = Decimal(value)
        if self._current_decimal > self._maximum_decimal:
            self._current_decimal = self._maximum_decimal
        if self._current_decimal < self._minimum_decimal:
            self._current_decimal = self._minimum_decimal
        self.decimalValueChanged.emit(self._current_decimal)


class DecimalSpinBoxWithButton(DecimalSpinBox):
    """Decimal spin box with button."""

    buttonClicked = QtCore.Signal()  # noqa: N815

    def __init__(
        self,
        button_text: str,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)
        self.button = QtWidgets.QPushButton(button_text, self)
        self.button.setFixedWidth(50)
        self.button.setObjectName("spinButton")
        self.button.clicked.connect(self.buttonClicked.emit)

        self.setButtonSymbols(QtWidgets.QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.setMinimum(Decimal(0))

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        """Adjust button position on resize.

        Button will be placed on the right side of the spin box with a distance
        of 5% of the border.
        """
        super().resizeEvent(event)
        self.update_button_position()

    def update_button_position(self) -> None:
        """Update button position."""
        button_size = self.button.size()
        self.button.move(
            self.width() - button_size.width() - 10,
            (self.height() - button_size.height()) // 2,
        )

    def show(self) -> None:
        """Override show method to update button position."""
        super().show()
        self.update_button_position()



"""Dialog to set manage configurations."""

from typing import Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Signal
from PySide6.QtGui import QPixmap

from plutus_terminal.ui.widgets.config.account_config import AccountConfig
from plutus_terminal.ui.widgets.config.news_config import NewsConfig
from plutus_terminal.ui.widgets.config.perps_config import PerpsConfig
from plutus_terminal.ui.widgets.config.web3_config import Web3Config


class ConfigDialog(QtWidgets.QDialog):
    """Config dialog."""

    updated_trade_values = Signal()
    leverage_changed = Signal(int)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize dialog."""
        super().__init__(parent)

        self._main_layout = QtWidgets.QVBoxLayout()
        self._main_layout.setContentsMargins(0, 0, 0, 0)

        self._tab_widget = QtWidgets.QTabWidget()
        self.persp_config = PerpsConfig()
        self.news_config = NewsConfig()
        self.web3_config = Web3Config()
        self.account_config = AccountConfig()

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Config widgets."""
        self.setWindowTitle("Plutus Terminal - Configuration")
        self.setWindowIcon(QPixmap(":/icons/plutus_icon"))
        self.setModal(False)
        self.setMinimumSize(800, 800)

        self._setup_persp_config()
        self.persp_config.updated_trade_values.connect(self.updated_trade_values)
        self.persp_config.leverage_changed.connect(self.leverage_changed)
        self._tab_widget.addTab(self.persp_config, "Trade")

        self._tab_widget.addTab(self.news_config, "News Source")

        self._tab_widget.addTab(self.web3_config, "Web3")

        self._tab_widget.addTab(self.account_config, "Account")

    def _setup_persp_config(self) -> None:
        """Config persp config."""
        self.persp_config.top_bar.hide()

    def _setup_layout(self) -> None:
        """Config layout."""
        self._main_layout.addWidget(self._tab_widget)
        self.setLayout(self._main_layout)



"""Spin BDecimal values."""

from typing import Optional

from PySide6 import QtCore, QtGui, QtWidgets


class DoubleSpinBoxWithButton(QtWidgets.QDoubleSpinBox):
    """Double spin box with button."""

    buttonClicked = QtCore.Signal()  # noqa: N815

    def __init__(
        self,
        button_text: str,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)
        self.button = QtWidgets.QPushButton(button_text, self)
        self.button.setFixedWidth(50)
        self.button.setObjectName("spinButton")
        self.button.clicked.connect(self.buttonClicked.emit)

        self.setButtonSymbols(QtWidgets.QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.setMinimum(0)
        self.setMaximum(1_000_000_000)
        self.setDecimals(8)
        self.setGroupSeparatorShown(True)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        """Adjust button position on resize.

        Button will be placed on the right side of the spin box with a distance
        of 5% of the border.
        """
        super().resizeEvent(event)
        self.update_button_position()

    def update_button_position(self) -> None:
        """Update button position."""
        button_size = self.button.size()
        self.button.move(
            self.width() - button_size.width() - 10,
            (self.height() - button_size.height()) // 2,
        )

    def show(self) -> None:
        """Override show method to update button position."""
        super().show()
        self.update_button_position()



"""Widget to display news data and interact with news."""

from __future__ import annotations

import asyncio
from functools import partial
import time
from typing import TYPE_CHECKING, Optional

from PySide6 import QtWidgets
from PySide6.QtCore import QSize, Qt, QTimer, QUrl, Signal
from PySide6.QtGui import QDesktopServices, QMouseEvent, QPixmap, QPixmapCache
from PySide6.QtNetwork import QNetworkAccessManager, QNetworkReply, QNetworkRequest
import re2  # type: ignore

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.exceptions import InvalidOrderSizeError
from plutus_terminal.core.exchange.types import PerpsTradeType
from plutus_terminal.core.types_ import NewsData, PerpsTradeDirection
from plutus_terminal.ui import ui_utils
from plutus_terminal.ui.widgets.image_web_viewer import ImageWebViewer
from plutus_terminal.ui.widgets.toast import Toast, ToastType

if TYPE_CHECKING:
    from collections.abc import Callable
    from decimal import Decimal

    from plutus_terminal.core.exchange.base import ExchangeBase, ExchangeFetcher

ICON_MAP = {
    "blogs": ":/sources/blog",
    "usgov": ":/sources/usgov",
    "binance en": ":/sources/binance",
    "upbit": ":/sources/upbit",
    "telegram": ":/sources/telegram",
    "crypto": ":/sources/crypto",
    "webs": ":/sources/webs",
    "medium": ":/sources/medium",
    "terminal": ":/sources/terminal",
}

NEWS_TIME_COLORS = {
    "green": 10,
    "yellow": 20,
}


class NewsWidget(QtWidgets.QGroupBox):
    """Widget to display news data and interact with news."""

    news_clicked = Signal(object)
    pair_clicked = Signal(str)
    timer_end = Signal()

    def __init__(
        self,
        news_data: NewsData,
        format_to_pair: Callable,
        available_pairs: set,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize shared variables."""
        super().__init__(parent=parent)
        self.news_data = news_data
        self.format_to_pair = format_to_pair
        self._available_pairs = available_pairs
        self._show_images = True
        self._re_percent_complied = re2.compile(r"\(([^)]+)%\)")
        self._async_tasks: list[asyncio.Task] = []

        self._max_time = 60
        self._elapsed_time = 0
        self._price_change: dict[str, str] = {}
        self._initial_prices: dict[str, Decimal] = {}
        self._icon_scale = 50

        self.main_layout = QtWidgets.QHBoxLayout()
        self.icon_layout = QtWidgets.QVBoxLayout()
        self.icon_label = QtWidgets.QLabel()
        self.info_layout = QtWidgets.QVBoxLayout()
        self.title_layout = QtWidgets.QHBoxLayout()
        self.title_label = QtWidgets.QLabel()
        self.timer = QTimer(self)
        self.stop_watch_label = QtWidgets.QLabel()
        self.body_frame = QtWidgets.QFrame()
        self.body_layout = QtWidgets.QVBoxLayout()
        self.body_label = QtWidgets.QLabel()
        self.body_image = ImageWebViewer()
        self.quote_icon_up = QtWidgets.QLabel()
        self.quote_frame = QtWidgets.QFrame()
        self.quote_layout = QtWidgets.QVBoxLayout()
        self.quote_title_layout = QtWidgets.QHBoxLayout()
        self.quote_title = QtWidgets.QLabel()
        self.quote_label = QtWidgets.QLabel()
        self.quote_image = ImageWebViewer()
        self.metadata_layout = QtWidgets.QHBoxLayout()
        self.time_label = QtWidgets.QLabel()
        self.feed_label = QtWidgets.QLabel()
        self.source_label = QtWidgets.QLabel()
        self.link_button = QtWidgets.QPushButton()
        self.coin_label = QtWidgets.QLabel()
        self.interactions_layout = QtWidgets.QVBoxLayout()

        self.group_box_layout: dict[str, QtWidgets.QVBoxLayout] = {}
        self.percent_label: dict[str, QtWidgets.QLabel] = {}

        self._setup_widgets()
        self._set_news_icon()
        self._setup_layout()

    @property
    def show_images(self) -> bool:
        """Return if images should be shown."""
        return self._show_images

    @show_images.setter
    def show_images(self, value: bool) -> None:
        """Set if images should be shown."""
        self._show_images = value
        if self.news_data["image"]:
            self.body_image.setVisible(value)
        if self.news_data["quote_image"]:
            self.quote_image.setVisible(value)

    def _setup_widgets(self) -> None:  # noqa: PLR0915
        """Create internal widgets."""
        self.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.icon_label.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.icon_label.setFixedSize(self._icon_scale, self._icon_scale)
        self.icon_label.setScaledContents(True)
        self.icon_label.setObjectName("newsIcon")

        self.title_label.setObjectName("title")
        self.title_label.setText(self.news_data["title"])
        self.title_label.setWordWrap(True)
        self.title_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse,
        )

        self.stop_watch_label.hide()
        self.stop_watch_label.setObjectName("newsStopWatch")
        self.stop_watch_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.stop_watch_label.setMaximumWidth(50)
        self.stop_watch_label.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )

        self.body_frame.setFrameStyle(QtWidgets.QFrame.Shape.Box)
        self.body_frame.setObjectName("newsFrameBody")

        if self.news_data["body"] or self.news_data["image"]:
            self.body_label.setObjectName("newsBody")
            self.body_label.setTextFormat(Qt.TextFormat.RichText)
            self.body_label.setText(self.news_data["body"])
            self.body_label.setWordWrap(True)
            self.body_label.setTextInteractionFlags(
                Qt.TextInteractionFlag.TextSelectableByMouse,
            )

            if self.news_data["image"]:
                self.body_image.set_image(self.news_data["image"])
            else:
                self.body_image.deleteLater()
        else:
            self.body_frame.hide()
            self.body_label.hide()

        if self.news_data["is_quote"] and any(
            (self.news_data["quote"], self.news_data["quote_image"]),
        ):
            quote_up_pixmap = QPixmap()
            QPixmapCache.find("quotes_up", quote_up_pixmap)
            if not quote_up_pixmap:
                quote_up_pixmap = QPixmap(":/icons/double_quote_up")
                QPixmapCache.insert("quotes_up", quote_up_pixmap)
            self.quote_icon_up.setPixmap(quote_up_pixmap)
            self.quote_icon_up.setAlignment(Qt.AlignmentFlag.AlignLeft)

            self.quote_frame.setFrameStyle(QtWidgets.QFrame.Shape.Box)
            self.quote_frame.setObjectName("newsFrameQuote")

            self.quote_title_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
            self.quote_title.setObjectName("subTitle")
            self.quote_title.setText(self.news_data["quoter"])

            self.quote_label.setObjectName("newsQuote")
            self.quote_label.setTextFormat(Qt.TextFormat.RichText)
            self.quote_label.setText(self.news_data["quote"])
            self.quote_label.setWordWrap(True)
            self.quote_label.setTextInteractionFlags(
                Qt.TextInteractionFlag.TextSelectableByMouse,
            )

            if self.news_data["quote_image"]:
                self.quote_image.set_image(self.news_data["quote_image"])
            else:
                self.quote_image.deleteLater()
        else:
            self.quote_frame.hide()

        self.link_button.setObjectName("newsLink")
        self.link_button.setFlat(True)
        self.link_button.setIcon(QPixmap(":/icons/external_link"))
        self.link_button.setIconSize(QSize(25, 25))
        self.link_button.setFixedSize(25, 25)
        self.link_button.clicked.connect(self.open_link)

        self.time_label.setObjectName("newsTime")
        self.time_label.setText(
            f'Source Time: {self.news_data["time"].strftime("%H:%M:%S:%f")}',
        )
        self.time_label.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignBottom,
        )
        self.time_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse,
        )

        self.feed_label.setObjectName("newsFeed")
        self.feed_label.setText(f"Feed: {self.news_data['feed']}")
        self.feed_label.setAlignment(
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignBottom,
        )
        self.feed_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse,
        )

        self.source_label.setObjectName("newsFeed")
        self.source_label.setText(f"Source: {self.news_data['source']}")
        self.source_label.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignBottom,
        )
        self.source_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse,
        )

        if self.news_data["coin"]:
            self.coin_label.setObjectName("newsCoin")
            self.coin_label.setText(f"Suggestions: {', '.join(self.news_data['coin'])}")

        self.setFixedWidth(500)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )

    def _set_news_icon(self) -> None:
        """Set icon to news, fetch from the internet if source is twitter."""
        source = self.news_data["source"].lower()
        if source == "twitter":
            icon_pixmap = QPixmap()
            # Try to get from cache if available
            QPixmapCache.find(self.news_data["icon"], icon_pixmap)

            # If not in cache, set temp icon and fetch from internet
            if not icon_pixmap:
                icon_pixmap = QPixmap(":/icons/no_token")
                network_manager = QNetworkAccessManager(self)
                network_manager.finished.connect(
                    partial(self._set_network_icon, self.icon_label),
                )
                network_manager.get(QNetworkRequest(QUrl(self.news_data["icon"])))
        else:
            icon_pixmap = QPixmap()
            QPixmapCache.find(source, icon_pixmap)
            if not icon_pixmap:
                icon_pixmap = QPixmap(ICON_MAP.get(source, ":/icons/no_token"))
                QPixmapCache.insert(source, icon_pixmap)

        self.icon_label.setPixmap(icon_pixmap)

    def _set_network_icon(self, target_label: QtWidgets.QLabel, reply: QNetworkReply) -> None:
        """Set icon pixmap from network reply."""
        image_data = reply.readAll()
        pixmap = QPixmap()
        pixmap.loadFromData(image_data)
        QPixmapCache.insert(self.news_data["icon"], pixmap)
        target_label.setPixmap(pixmap)

    def _setup_layout(self) -> None:
        """Connect widgets to layouts."""
        self.main_layout.addLayout(self.icon_layout)
        self.icon_layout.addWidget(self.icon_label)
        self.icon_layout.addWidget(self.stop_watch_label)
        self.icon_layout.addStretch()

        self.main_layout.addLayout(self.info_layout)
        self.title_layout.addWidget(self.title_label)
        self.title_layout.addWidget(self.time_label)
        self.info_layout.addLayout(self.title_layout)

        self.body_frame.setLayout(self.body_layout)
        self.body_layout.addWidget(self.body_label)
        self.body_layout.addWidget(self.body_image)
        self.body_layout.addWidget(self.quote_label)
        self.body_layout.addWidget(self.quote_frame)

        self.quote_frame.setLayout(self.quote_layout)
        self.quote_title_layout.addWidget(self.quote_icon_up)
        self.quote_title_layout.addWidget(self.quote_title)
        self.quote_layout.addLayout(self.quote_title_layout)
        self.quote_layout.addWidget(self.quote_label)
        self.quote_layout.addWidget(self.quote_image)

        self.info_layout.addWidget(self.body_frame)
        self.metadata_layout.addWidget(self.feed_label)
        self.metadata_layout.addWidget(self.source_label)
        self.metadata_layout.addWidget(self.link_button)
        self.info_layout.addLayout(self.metadata_layout)
        self.info_layout.addWidget(self.coin_label)

        self.info_layout.addLayout(self.interactions_layout)

        self.setLayout(self.main_layout)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Emit signal when clicked."""
        if event.button() == Qt.MouseButton.LeftButton:
            self.news_clicked.emit(self)
        return super().mousePressEvent(event)

    async def set_initial_prices(self, fetch_price_at_time: Callable) -> None:
        """Set initial prices for coins in data."""
        for coin in self.news_data["coin"]:
            pair = self.format_to_pair(coin)
            if pair in self._available_pairs:
                current_price = await fetch_price_at_time(
                    pair,
                    self.news_data["time"].timestamp(),
                )
                current_price = current_price["price"]
                self._initial_prices[pair] = current_price
                minimal_digits = ui_utils.get_minimal_digits(current_price, 4)
                initial_price_label = QtWidgets.QLabel(
                    f"Price at news: {current_price:,.{minimal_digits}f}",
                )
                percent_layout = self.group_box_layout[coin].itemAt(0).layout()
                percent_layout.insertWidget(1, initial_price_label)  # type: ignore

    def create_interactions(self, exchange: ExchangeBase) -> bool:  # noqa: PLR0915
        """Create buttons for interactions.

        Args:
            exchange (ExchangeBase): Exchange instance.

        Returns:
            bool: If an interaction was created.
        """
        interaction_created = False
        interaction_pairs = set()
        for coin in self.news_data["coin"]:
            pair = self.format_to_pair(coin)

            if pair not in self._available_pairs:
                continue
            interaction_pairs.add(pair)

            button_group_box = ClickableGroupBox(coin)
            button_group_box.clicked.connect(partial(self.pair_clicked.emit, pair))
            self.group_box_layout[coin] = QtWidgets.QVBoxLayout()
            self.percent_label[pair] = QtWidgets.QLabel("0%")
            self.percent_label[pair].hide()
            button_layout = QtWidgets.QGridLayout()

            percent_layout = QtWidgets.QHBoxLayout()
            percent_layout.addWidget(self.percent_label[pair])
            self.group_box_layout[coin].addLayout(percent_layout)
            self.group_box_layout[coin].addLayout(button_layout)
            button_group_box.setLayout(self.group_box_layout[coin])

            option_keys = (
                "trade_value_lowest",
                "trade_value_low",
                "trade_value_medium",
                "trade_value_high",
            )
            for index, option_key in enumerate(option_keys):
                value = getattr(
                    CONFIG,
                    option_key,
                )
                button_long = QtWidgets.QPushButton(f"${value}")
                button_long.setObjectName(f"LONG_{index}")
                button_long.setMinimumHeight(25)
                button_long.clicked.connect(
                    partial(
                        self.handle_interaction_click,
                        exchange.create_order,
                        pair,
                        option_key,
                        PerpsTradeDirection.LONG,
                        PerpsTradeType.MARKET,
                    ),
                )
                button_long.setProperty("class", "LONG")
                match index:
                    case 0:
                        button_layout.addWidget(button_long, 0, 0)
                    case 1:
                        button_layout.addWidget(button_long, 0, 1)
                    case 2:
                        button_layout.addWidget(button_long, 1, 0)
                    case 3:
                        button_layout.addWidget(button_long, 1, 1)

            for index, option_key in enumerate(option_keys):
                value = getattr(
                    CONFIG,
                    option_key,
                )
                button_short = QtWidgets.QPushButton(f"-${value}")
                button_short.setObjectName(f"SHORT_{index}")
                button_short.setMinimumHeight(25)
                button_short.clicked.connect(
                    partial(
                        self.handle_interaction_click,
                        exchange.create_order,
                        pair,
                        option_key,
                        PerpsTradeDirection.SHORT,
                        PerpsTradeType.MARKET,
                    ),
                )
                button_short.setProperty("class", "SHORT")
                match index:
                    case 0:
                        button_layout.addWidget(button_short, 0, 2)
                    case 1:
                        button_layout.addWidget(button_short, 0, 3)
                    case 2:
                        button_layout.addWidget(button_short, 1, 2)
                    case 3:
                        button_layout.addWidget(button_short, 1, 3)

            self.interactions_layout.addWidget(button_group_box)
            interaction_created = True

        if interaction_created:
            # Set price of news at source time
            self._async_tasks.append(
                asyncio.create_task(
                    self.set_initial_prices(exchange.fetcher.fetch_price_at_time),
                ),
            )

            if self.news_data["time"].timestamp() > time.time() - 60:
                for pair in interaction_pairs:
                    self.percent_label[pair].show()
                    self._async_tasks.append(
                        asyncio.create_task(exchange.fetcher.subscribe_to_price(pair)),
                    )

                exchange.fetcher_bus.subscribed_prices_signal.connect(
                    self.update_percents,
                )
                self.timer_end.connect(
                    partial(
                        exchange.fetcher_bus.subscribed_prices_signal.disconnect,
                        self.update_percents,
                    ),
                )
                self.timer_end.connect(
                    partial(
                        self._unsubscribe_from_price_updates,
                        interaction_pairs,
                        exchange.fetcher,
                    ),
                )

                self.start_timer()
        return interaction_created

    def handle_interaction_click(
        self,
        trade_function: Callable,
        coin: str,
        config_key_value: str,
        trade_direction: PerpsTradeDirection,
        trade_type: PerpsTradeType,
    ) -> None:
        """Handle buys/sells from interaction buttons.

        This will ensure that the amount is updated dynamically at button press.

        Args:
            trade_function (Callable): Trade function.
            coin (str): Coin to trade.
            config_key_value (str): Config key value to get amount from.
            trade_direction (PerpsTradeDirection): Trade direction.
            trade_type (PerpsTradeType): Trade type.
        """
        amount = getattr(CONFIG, config_key_value)
        try:
            trade_function(coin, amount, trade_direction, trade_type)
        except InvalidOrderSizeError as error:
            Toast.show_message(
                f"{error}",
                type_=ToastType.ERROR,
            )

    def _unsubscribe_from_price_updates(
        self,
        pairs: set[str],
        exchange_fetcher: ExchangeFetcher,
    ) -> None:
        """Unsubscribe from price updates.

        Args:
            pairs (set[str]): Pairs to unsubscribe.
            exchange_fetcher (ExchangeFetcher): Exchange fetcher.
        """
        for pair in pairs:
            self._async_tasks.append(
                asyncio.create_task(exchange_fetcher.unsubscribe_to_price(pair)),
            )

    def start_timer(self) -> None:
        """Start time for each 1 second."""
        self.timer.timeout.connect(self._update_on_timer)
        self.stop_watch_label.setVisible(True)
        self.timer.start(1000)

    def _update_on_timer(self) -> None:
        """Update label with time."""
        self._elapsed_time += 1
        if self._elapsed_time < NEWS_TIME_COLORS["green"]:
            self.stop_watch_label.setProperty("class", "success")
            self.stop_watch_label.style().polish(self.stop_watch_label)
        elif self._elapsed_time < NEWS_TIME_COLORS["yellow"]:
            self.stop_watch_label.setProperty("class", "warning")
            self.stop_watch_label.style().polish(self.stop_watch_label)
        else:
            self.stop_watch_label.setProperty("class", "danger")
            self.stop_watch_label.style().polish(self.stop_watch_label)

        self.stop_watch_label.setText(str(self._elapsed_time))
        for pair in self._initial_prices:
            try:
                price_change = self._price_change.get(pair, "(0.00%)")
                percent_change = self._re_percent_complied.search(price_change)
                # Slice to get digits only
                if float(percent_change.group(1)) > 0:  # type: ignore
                    self.percent_label[pair].setStyleSheet("color: rgb(100, 255, 100);")
                else:
                    self.percent_label[pair].setStyleSheet("color: red;")
                self.percent_label[pair].setText(price_change)
            except KeyError:
                continue

        if self._elapsed_time >= self._max_time:
            self.timer.stop()
            self.stop_watch_label.setVisible(False)
            for pair in self._initial_prices:
                try:
                    self.percent_label[pair].hide()
                except KeyError:
                    continue
            self.timer_end.emit()

    def update_percents(self, cached_prices: dict) -> None:
        """Update percents labels for tokens."""
        for pair, initial_price in self._initial_prices.items():
            # In case the websocket reply is a bit late
            pair_data = cached_prices.get(pair, {"price": initial_price})
            current_price = pair_data["price"]
            percentage = ((current_price / initial_price) - 1) * 100
            minimal_digits = ui_utils.get_minimal_digits(current_price, 4)
            self._price_change[pair] = (
                f"{current_price:,.{minimal_digits}f} ({round(percentage, 3):.3f}%)"
            )

    def set_selected_style(self) -> None:
        """Set border to selected style."""
        self.setProperty("class", "selected")
        self.style().polish(self)

    def set_unselected_style(self) -> None:
        """Set border to unselected style."""
        self.setProperty("class", "unselected")
        self.style().polish(self)

    def open_link(self) -> None:
        """Open link in browser."""
        QDesktopServices.openUrl(self.news_data["link"])

    def update_trade_buttons(self) -> None:
        """Update trade values."""
        value_map = {
            0: CONFIG.trade_value_lowest,
            1: CONFIG.trade_value_low,
            2: CONFIG.trade_value_medium,
            3: CONFIG.trade_value_high,
        }
        for index in range(4):
            for widget in self.findChildren(QtWidgets.QPushButton, f"SHORT_{index}"):
                widget.setText(f"-${value_map[index]}")
            for widget in self.findChildren(QtWidgets.QPushButton, f"LONG_{index}"):
                widget.setText(f"${value_map[index]}")


class ClickableGroupBox(QtWidgets.QGroupBox):
    """GroupBox clickable."""

    clicked = Signal()

    def __init__(self, title: str, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize widget."""
        super().__init__(title, parent=parent)
        self.installEventFilter(self)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setMouseTracking(True)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Emit signal when clicked."""
        self.clicked.emit()
        return super().mousePressEvent(event)



"""Types for news."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from datetime import datetime


class NewsData(TypedDict):
    """News data dict from news source."""

    title: str
    link: str
    body: str
    image: str
    is_quote: bool
    quote: str
    quoter: str
    quote_image: str
    icon: str
    source: str
    time: datetime
    coin: set[str]
    feed: str
    sfx: str
    ignored: bool



"""Top bar widget."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from PySide6 import QtWidgets
from PySide6.QtGui import QPixmap

from plutus_terminal.ui.widgets.account_picker import AccountPicker
from plutus_terminal.ui.widgets.clock import WebClock

if TYPE_CHECKING:
    from plutus_terminal.ui.widgets.config import ConfigDialog


class UserTopBar(QtWidgets.QWidget):
    """Top bar widget."""

    def __init__(
        self,
        config_dialog: ConfigDialog,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent)
        self.setObjectName("user_top_bar")
        self.main_layout = QtWidgets.QVBoxLayout()
        self._bar_layout = QtWidgets.QHBoxLayout()

        self.account_picker = AccountPicker()
        self._clock = WebClock()
        self._config_dialog = config_dialog
        self._config_button = QtWidgets.QPushButton()
        self._config_button.setIcon(QPixmap(":/icons/config_icon"))
        self._config_button.setMinimumSize(30, 30)
        self._config_button.setProperty("class", "borderless")
        self._config_button.clicked.connect(self._config_dialog.show)
        self._line = QtWidgets.QFrame()
        self._line.setObjectName("divisor")
        self._line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self._line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        self._bar_layout.addWidget(self.account_picker)
        self._bar_layout.addStretch()
        self._bar_layout.addWidget(self._clock)
        self._bar_layout.addStretch()
        self._bar_layout.addWidget(self._config_button)
        self.main_layout.addLayout(self._bar_layout)
        self.main_layout.addWidget(self._line)

        self.setLayout(self.main_layout)



"""Toaster Widget."""

from enum import Enum
import logging
import os
from typing import ClassVar, Optional

from PySide6.QtCore import (
    QEvent,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTimer,
    Signal,
)
from PySide6.QtGui import (
    QCloseEvent,
    QCursor,
    QPixmap,
    QResizeEvent,
    QShowEvent,
)
from PySide6.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

LOGGER = logging.getLogger(__name__)


class ToastType(Enum):
    """Toaster type enum."""

    MESSAGE = "message"
    SUCCESS = "success"
    WARNING = "warning"
    ERROR = "error"


class Toast(QFrame):
    """Toast Message widget for top of screen."""

    _toasts_win: ClassVar[dict] = {}
    _toasts_desktop: ClassVar[dict] = {}
    closed = Signal()

    def __init__(self, parent: Optional[QWidget] = None, desktop: bool = False) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)
        self.parent_rect = QRect()
        self._desktop = desktop

        self._main_layout = QVBoxLayout(self)
        self._close_button = QToolButton()
        self._terminal_button = QToolButton()
        self._desktop_label = QLabel()
        self._pin_button = QToolButton()
        self._button_layout = QHBoxLayout()
        self._message_widget = QWidget()
        self._id = os.urandom(8)

        self._timer = QTimer()
        self._margin = 10
        self._toast_distance = 5

        self._setup_widgets()
        self._setup_layout()

    def add_message_widget(self, message_widget: QWidget) -> None:
        """Add message widget."""
        self._message_widget = message_widget
        self._main_layout.addWidget(self._message_widget)

    def _setup_widgets(self) -> None:
        """Configure internal widgets."""
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.setWindowFlags(
            Qt.WindowType.FramelessWindowHint
            | Qt.WindowType.BypassWindowManagerHint
            | Qt.WindowType.WindowStaysOnTopHint,
        )

        if self.parent():
            # we have a parent, install an eventFilter so that when it's resized
            # the notification will be correctly moved to the right corner
            self.parent().installEventFilter(self)

        self._pin_button.setIcon(QPixmap(":/icons/pushpin_icon"))
        self._pin_button.setAutoRaise(True)
        self._pin_button.setToolTip("Pin toask and prevent to auto close")
        self._pin_button.clicked.connect(self._pin_toast)

        self._close_button.setIcon(QPixmap(":/icons/close_icon"))
        self._close_button.setAutoRaise(True)
        self._close_button.setToolTip("Close toast")
        self._close_button.clicked.connect(self.close)

        if self._desktop:
            self._terminal_button.setIcon(QPixmap(":/icons/terminal_icon"))
            self._terminal_button.setAutoRaise(True)
            self._terminal_button.setToolTip("Open plutus terminal")
            self._terminal_button.clicked.connect(self._open_main_window)

            self._desktop_label.setText("Plutus Terminal")
            self._desktop_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
            self._desktop_label.setObjectName("title")

        self._timer.timeout.connect(self.close)

    def _setup_layout(self) -> None:
        """Configure layout."""
        if self._desktop:
            self._button_layout.addWidget(self._terminal_button)
            self._button_layout.addStretch()
            self._button_layout.addWidget(self._desktop_label)

        self._button_layout.addStretch()
        self._button_layout.addWidget(self._pin_button)
        self._button_layout.addWidget(self._close_button)
        self._main_layout.addLayout(self._button_layout)

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        """Filter resize events.

        Args:
            watched: The object being watched
            event: The event being watched

        Returns:
            bool: True if the event was handled.
        """
        if watched == self.parent() and event.type() == QEvent.Type.Resize:
            self.parent_rect = self.parent().rect()  # type: ignore
            self.adjust_position()
        return super().eventFilter(watched, event)

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Adjust position on resize."""
        self.adjust_position()
        return super().resizeEvent(event)

    def closeEvent(self, event: QCloseEvent) -> None:
        """Delete widget on close."""
        self._timer.stop()
        try:
            if self._desktop:
                del Toast._toasts_desktop[self._id]
            else:
                del Toast._toasts_win[self._id]
        except KeyError:
            LOGGER.warning("Toast %s not found", self._id)

        event.accept()

        existent_toasts = (
            list(Toast._toasts_desktop.values())
            if self._desktop
            else list(Toast._toasts_win.values())
        )

        for toast in existent_toasts:
            toast.adjust_position()

    def showEvent(self, event: QShowEvent) -> None:
        """Start time on show."""
        self.adjustSize()
        self.adjust_position(add_event=True)
        self._timer.start()
        self.raise_()
        return super().showEvent(event)

    def adjust_position(self, add_event: bool = False) -> None:
        """Move the toast to corner of parent rect and adjust for multiple messages."""
        geometry = self.geometry()

        offset = 0

        existent_toasts = (
            list(Toast._toasts_desktop.values())
            if self._desktop
            else list(Toast._toasts_win.values())
        )

        toasts = reversed(existent_toasts) if add_event else existent_toasts

        for toast in toasts:
            if toast is not self:
                offset += toast.geometry().height() + self._toast_distance
            elif not add_event and self:
                break

        geometry.moveBottomRight(
            self.parent_rect.bottomRight() + QPoint(-self._margin, -self._margin - offset),
        )

        self.setGeometry(geometry)

    def _open_main_window(self) -> None:
        """Open main window."""
        self.close()
        for widget in QApplication.topLevelWidgets():
            if isinstance(widget, QMainWindow):
                main_window = widget
                main_window.show()

    def _pin_toast(self) -> None:
        """Pin toast and prevent to auto close."""
        if self._timer.isActive():
            self.setObjectName("pinned")
            self._pin_button.setIcon(QPixmap(":/icons/unpin_icon"))
            self._timer.stop()
        else:
            self.setObjectName("")
            self._pin_button.setIcon(QPixmap(":/icons/pushpin_icon"))
            self._timer.start()
        self.style().polish(self)

    @staticmethod
    def show_message(
        message: str,
        timeout: int = 10000,
        desktop: bool = False,
        type_: ToastType = ToastType.MESSAGE,
    ) -> bytes:
        """Show Toast message.

        Args:
            message (str): Message to show.
            timeout (int, optional): Timeout in milliseconds. Defaults to 10000.
            desktop (bool, optional): Show on desktop. Defaults to False.
            type_ (ToastType, optional): Toast type. Defaults to ToastType.MESSAGE.

        Returns:
            bytes: Toast ID.
        """
        message_label = QLabel(message)
        message_label.setTextFormat(Qt.TextFormat.RichText)
        message_label.setOpenExternalLinks(True)
        return Toast.show_widget(message_label, timeout, desktop, type_)

    @staticmethod
    def update_message(
        toast_id: bytes,
        message: str,
        type_: ToastType,
        desktop: bool = False,
    ) -> None:
        """Update toast message.

        Args:
            toast_id (bytes): Toast ID.
            message (str): Message to show.
            type_ (ToastType): Toast type.
            desktop (bool, optional): Show on desktop. Defaults to False.
        """
        if desktop:
            toast = Toast._toasts_desktop.get(toast_id, False)
        else:
            toast = Toast._toasts_win.get(toast_id, False)

        if not toast:
            return
        toast._message_widget.setText(message)  # noqa: SLF001
        toast._timer.start()  # noqa: SLF001
        toast.setProperty("class", type_.value)
        toast.style().polish(toast)
        toast.resize(toast.sizeHint())
        toast.adjust_position()

    @staticmethod
    def show_widget(
        message_widget: QWidget,
        timeout: int = 10000,
        desktop: bool = False,
        type_: ToastType = ToastType.MESSAGE,
    ) -> bytes:
        """Show toast widget.

        Args:
            message_widget (QWidget): Message widget to show.
            timeout (int, optional): Timeout in milliseconds. Defaults to 10000.
            desktop (bool, optional): Show on desktop. Defaults to False.
            type_ (ToastType, optional): Toast type. Defaults to ToastType.MESSAGE.

        Returns:
            bytes: Toast ID.
        """
        if desktop:
            # Setup to show on desktop
            toast = Toast(None, desktop=True)
            current_screen = QApplication.primaryScreen()

            # Use cursor for window reference
            reference = QRect(QCursor.pos() - QPoint(1, 1), QSize(3, 3))

            max_area = 0
            for screen in QApplication.screens():
                screen_rect = screen.geometry().intersected(reference)
                area = screen_rect.width() * screen_rect.height()
                if area > max_area:
                    max_area = area
                    current_screen = screen
                    break
            parent_rect = current_screen.availableGeometry()
        else:
            # Setup to show on top of main window
            main_window = None
            for widget in QApplication.topLevelWidgets():
                if isinstance(widget, QMainWindow):
                    main_window = widget
                    break
            if not main_window:
                msg = "No main window found"
                raise ValueError(msg)
            parent = main_window
            toast = Toast(parent)
            parent_rect = parent.rect()

        # Store toast in memory depending of type
        if desktop:
            Toast._toasts_desktop[toast._id] = toast  # noqa: SLF001
        else:
            Toast._toasts_win[toast._id] = toast  # noqa: SLF001

        toast._timer.setInterval(timeout)  # noqa: SLF001

        # Add message to toast
        toast.add_message_widget(message_widget)
        toast.setProperty("class", type_.value)

        toast.parent_rect = parent_rect
        # Extra flag to prevent taskbar icon
        if desktop:
            toast.setWindowFlag(Qt.WindowType.Tool)
        toast.show()
        return toast._id  # noqa: SLF001



"""Handle news from Tree Of Alpha."""

from __future__ import annotations

from datetime import datetime, timezone
import logging
from typing import TYPE_CHECKING, Optional

from httpx import AsyncClient
import keyring
import orjson as json
import re2
from tenacity import (
    before_sleep_log,
    retry,
    stop_after_attempt,
    wait_exponential,
)
from websockets.client import WebSocketClientProtocol, connect

from plutus_terminal.core.news.base import NewsFetcher
from plutus_terminal.core.types_ import NewsData
from plutus_terminal.log_utils import log_retry

if TYPE_CHECKING:
    from plutus_terminal.core.news.base import NewsMessageBus

LOGGER = logging.getLogger(__name__)

TREE_KEY_NAME = "TreeOfAlpha"


class TreeNews(NewsFetcher):
    """News fetcher for Tree Of Alpha News."""

    def __init__(self) -> None:
        """Initialize shared variables."""
        self.wss = "wss://news.treeofalpha.com/ws"
        self._socket: Optional[WebSocketClientProtocol] = None  # type: ignore
        self._compiled_pattern = re2.compile(r"\bQuote\s+\[(@\w+)\]\([^)]*\)")

    async def websocket_connect(self) -> WebSocketClientProtocol:
        """Connect to websocket to fetch prices.

        Returns:
            WebSocketClientProtocol: Websocket connection.
        """
        if self._socket is None or self._socket.closed:
            self._socket = await connect(self.wss, ping_interval=5, ping_timeout=10)
            LOGGER.info("Connected to TreeOfAlpha Websocket")
        return self._socket

    @retry(
        wait=wait_exponential(multiplier=1, min=0.4, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def _ensure_websocket_connection(self) -> None:
        """Ensure websocket is connected."""
        if self._socket is None or self._socket.closed:
            LOGGER.warning(
                "TreeOfAlpha Websocket disconnected. Attempting to reconnect websocket...",
            )
            await self.websocket_connect()

    @retry(
        wait=wait_exponential(multiplier=1, min=0.4, max=5),
        reraise=True,
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def subscribe_to_wss(self, message_bus: NewsMessageBus) -> None:
        """Subscribe to news wss and emit news signal on new entry.

        Args:
            message_bus (plutus_terminal.ui.thread.NewsMessageBus): Message bus
                to emit news messages
        """
        await self._ensure_websocket_connection()

        LOGGER.info("Subscribed to TreeOfAlpha news source.")
        async for message in self._socket:  # type: ignore
            LOGGER.debug("New raw message received from TreeOfAlpha")
            json_message = json.loads(message)
            formated_message = self.format_news(json_message)
            message_bus.raw_news_signal.emit(formated_message)

    async def login(self) -> None:
        """Login to news source."""
        LOGGER.info("Logging in to TreeOfAlpha...")
        tree_api_key = keyring.get_password(
            "plutus-terminal:news-source",
            TREE_KEY_NAME,
        )
        if not tree_api_key:
            LOGGER.warning("TreeOfAlph API key not found")
            return
        await self._ensure_websocket_connection()
        if not self._socket:
            return
        await self._socket.send(f"login {tree_api_key}")
        login_attempt = await self._socket.recv()
        login_attempt = json.loads(login_attempt)
        login_attempt.setdefault("user", {})
        login_attempt["user"].pop("address", None)
        LOGGER.info("TreeOfAlpha login result: %s", login_attempt)

    @retry(
        wait=wait_exponential(multiplier=1, min=0.4, max=2),
        stop=stop_after_attempt(5),
        reraise=True,
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_old_news(self, limit: int) -> list[NewsData]:
        """Fetch old news from API.

        Args:
            limit (int): Amount of news to fetch.

        Returns:
            list[NewsData]: List of old news. This list is espected to be ordered
            from latest to oldest.
        """
        request_url = f"https://news.treeofalpha.com/api/news?limit={limit}"
        async with AsyncClient() as client:
            response = await client.get(request_url)
        response.raise_for_status()
        data = response.json()
        list_news = [self.format_news(news) for news in data]
        return list_news[::-1]

    def format_news(self, news_message: dict) -> NewsData:
        """Format given news.

        Args:
            news_message (dict): News message to be formated.

        Returns:
            NewsData: Formated news.
        """
        title = news_message.get("en", news_message.get("title", ""))
        link = news_message.get("url", news_message.get("link", ""))
        body = news_message.get("body", "")
        icon = news_message.get("icon", "")
        source = news_message.get("source", news_message.get("type", ""))
        time = datetime.fromtimestamp(news_message["time"] / 1000, timezone.utc)
        coin = {news_message.get("coin", "")} if news_message.get("coin", "") else set()
        suggestions = news_message.get("suggestions", [])
        image = news_message.get("image", "")

        is_quote = False
        quote = ""
        quoter = ""
        quote_image = ""

        if suggestions:
            for item in suggestions:
                coin.add(item["coin"])

        if not body:
            title_split = title.split(":")
            title = title_split[0].strip()
            body = "".join(title_split[1:]).strip()

        if news_message.get("type") == "direct":
            source = "Twitter"

        if source == "Twitter":
            is_quote = news_message["info"]["isQuote"]

        if is_quote:
            match = self._compiled_pattern.search(body)
            if match:
                quote = body[match.end() :].strip()
                body = body[: match.start()].strip()
                quoter = str(match.group(1)).strip()

        return NewsData(
            title=title,
            link=link,
            body=body,
            image=image,
            is_quote=is_quote,
            quote=quote,
            quoter=quoter,
            quote_image=quote_image,
            icon=icon,
            source=source,
            time=time,
            coin=coin,
            feed="Tree Of Alpha",
            sfx=":/sfx/coin",
            ignored=False,
        )

    async def stop_async(self) -> None:
        """Stop infinite loops and close connections."""
        if self._socket is not None:
            await self._socket.close()



"""Widget to control news configuration."""

from __future__ import annotations

from functools import partial
from typing import Optional

import keyring
import orjson
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtMultimedia import QSoundEffect

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.db.models import UserFilter
from plutus_terminal.core.news.filter._actions import FILTER_ACTIONS_MAP
from plutus_terminal.core.news.filter.types import ActionType, FilterType
from plutus_terminal.core.news.phoenix_news import PHOENIX_KEY_NAME
from plutus_terminal.core.news.tree_news import TREE_KEY_NAME
from plutus_terminal.ui.ui_utils import list_resources_from_prefix
from plutus_terminal.ui.widgets.toast import Toast, ToastType
from plutus_terminal.ui.widgets.top_bar_widget import TopBar


class NewsConfig(QtWidgets.QWidget):
    """Widget to control news configuration."""

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        """Initialize shared attributes."""
        super().__init__(parent=parent)

        self._main_layout = QtWidgets.QVBoxLayout()

        self._news_source_bar = TopBar("News Source API Keys")
        self._tree_box = QtWidgets.QGroupBox("TreeOfAlpha API Key")
        self._tree_box_layout = QtWidgets.QVBoxLayout()
        self._tree_text_label = QtWidgets.QLabel()
        self._tree_input = QtWidgets.QLineEdit()
        self._phoenix_box = QtWidgets.QGroupBox("Phoenix API Key")
        self._phoenix_box_layout = QtWidgets.QVBoxLayout()
        self._phoenix_text_label = QtWidgets.QLabel()
        self._phoenix_input = QtWidgets.QLineEdit()

        self._news_filters = TopBar("News Filters")
        self._news_scroll_area = QtWidgets.QScrollArea()
        self._news_scroll_wdiget = QtWidgets.QWidget()
        self._news_scroll_layout = QtWidgets.QVBoxLayout()

        self._keyword_matching_layout = QtWidgets.QVBoxLayout()
        self._keyword_matching_box = QtWidgets.QGroupBox("Keyword Matching - Filter")
        self._keyword_matching_add_btn = QtWidgets.QPushButton("Add filter")

        self._data_matching_layout = QtWidgets.QVBoxLayout()
        self._data_matching_box = QtWidgets.QGroupBox("Data Matching - Filter")
        self._data_matching_add_btn = QtWidgets.QPushButton("Add filter")

        self._reset_filters_btn = QtWidgets.QPushButton("Reset Filters")
        self._save_filters_btn = QtWidgets.QPushButton("Save Filters")

        self._setup_widgets()
        self._setup_layout()

    def _setup_widgets(self) -> None:
        """Config widgets."""
        self._tree_text_label.setWordWrap(True)
        self._tree_text_label.setOpenExternalLinks(True)
        self._tree_text_label.setText(
            """Add your TreeOfAlpha API key bellow if you are a paid subscriber.<br>"""
            """To get your API key, go to """
            """<a href="https://news.treeofalpha.com/api/api_key """
            """style="color:rgb(80, 210, 180)">"""
            """https://news.treeofalpha.com/api/api_key</a>""",
        )
        current_tree_key = keyring.get_password(
            "plutus-terminal:news-source",
            TREE_KEY_NAME,
        )
        if current_tree_key:
            self._tree_input.setText(current_tree_key)
        else:
            self._tree_input.setPlaceholderText(
                "Enter your TreeOfAlpha API key here...",
            )

        self._phoenix_text_label.setWordWrap(True)
        self._phoenix_text_label.setText(
            """Add your PhoenixNews API key bellow if you are a paid subscriber.<br>"""
            """To get your API key, go to """
            """<a href="https://phoenixnews.io", style="color:rgb(80, 210, 180)">"""
            """https://phoenixnews.io</a>""",
        )
        self._phoenix_text_label.setOpenExternalLinks(True)
        current_phoenix_key = keyring.get_password(
            "plutus-terminal:news-source",
            PHOENIX_KEY_NAME,
        )
        if current_phoenix_key:
            self._phoenix_input.setText(current_phoenix_key)
        else:
            self._phoenix_input.setPlaceholderText("Enter your Phoenix API key here...")

        self._tree_input.editingFinished.connect(
            partial(self.record_news_source_key, TREE_KEY_NAME),
        )
        self._phoenix_input.editingFinished.connect(
            partial(self.record_news_source_key, PHOENIX_KEY_NAME),
        )

        self._news_scroll_area.setWidgetResizable(True)

        self._keyword_matching_add_btn.setMinimumSize(80, 30)
        self._keyword_matching_add_btn.setProperty("class", "LONG")
        self._keyword_matching_add_btn.clicked.connect(self._add_keyword_filter)

        self._data_matching_add_btn.setMinimumSize(80, 30)
        self._data_matching_add_btn.setProperty("class", "LONG")
        self._data_matching_add_btn.clicked.connect(self._add_data_filter)

        self._reset_filters_btn.setMinimumSize(80, 30)
        self._reset_filters_btn.clicked.connect(self._reset_filters)

        self._save_filters_btn.setMinimumSize(80, 30)
        self._save_filters_btn.setProperty("class", "LONG")
        self._save_filters_btn.clicked.connect(self._save_filters)

        user_filters = CONFIG.get_all_user_filters()
        for user_filter in user_filters:
            if int(user_filter.filter_type) == FilterType.KEYWORD_MATCHING:  # type: ignore
                self._keyword_matching_layout.addWidget(KeywordMatchingWidget(user_filter))
            if int(user_filter.filter_type) == FilterType.DATA_MATCHING:  # type: ignore
                self._data_matching_layout.addWidget(DataMatchingWidget(user_filter))

    def _setup_layout(self) -> None:
        """Config layout."""
        self._main_layout.addWidget(self._news_source_bar)
        self._tree_box_layout.addWidget(self._tree_text_label)
        self._tree_box_layout.addWidget(self._tree_input)
        self._tree_box.setLayout(self._tree_box_layout)
        self._main_layout.addWidget(self._tree_box)

        self._phoenix_box_layout.addWidget(self._phoenix_text_label)
        self._phoenix_box_layout.addWidget(self._phoenix_input)
        self._phoenix_box.setLayout(self._phoenix_box_layout)
        self._main_layout.addWidget(self._phoenix_box)

        self._main_layout.addWidget(self._news_filters)

        self._news_scroll_wdiget.setLayout(self._news_scroll_layout)
        self._news_scroll_area.setWidget(self._news_scroll_wdiget)

        self._keyword_matching_layout.addWidget(
            self._keyword_matching_add_btn,
            alignment=QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignCenter,
        )
        self._keyword_matching_box.setLayout(self._keyword_matching_layout)

        self._data_matching_layout.addWidget(
            self._data_matching_add_btn,
            alignment=QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignCenter,
        )
        self._data_matching_box.setLayout(self._data_matching_layout)

        self._news_scroll_layout.addWidget(self._keyword_matching_box)
        self._news_scroll_layout.addWidget(self._data_matching_box)
        filter_buttons_layout = QtWidgets.QHBoxLayout()
        filter_buttons_layout.addStretch()
        filter_buttons_layout.addWidget(self._reset_filters_btn)
        filter_buttons_layout.addWidget(self._save_filters_btn)
        self._news_scroll_layout.addLayout(filter_buttons_layout)
        self._news_scroll_layout.addStretch()

        self._main_layout.addWidget(self._news_scroll_area)
        self.setLayout(self._main_layout)

    def record_news_source_key(self, news_source: str) -> None:
        """Record the news source API key in keyring.

        Args:
            news_source: name of the news source,
        """
        text_source = {
            TREE_KEY_NAME: self._tree_input.text(),
            PHOENIX_KEY_NAME: self._phoenix_input.text(),
        }
        keyring.set_password(
            "plutus-terminal:news-source",
            news_source,
            text_source[news_source],
        )
        Toast.show_message(
            "News source API key saved successfully!",
            type_=ToastType.SUCCESS,
        )

    def _add_keyword_filter(self) -> None:
        """Add a new keyword filter."""
        user_filter = UserFilter(
            filter_type=FilterType.KEYWORD_MATCHING,
            match_pattern=orjson.dumps({"keyword": ""}).decode("utf-8"),
            action_type=ActionType.COIN_ASSOCIATION,
            action_args=orjson.dumps({"coin": "BTC", "color": [255, 0, 0]}).decode(
                "utf-8",
            ),
        )
        self._keyword_matching_layout.insertWidget(
            self._keyword_matching_layout.count() - 1,
            KeywordMatchingWidget(user_filter),
        )

    def _add_data_filter(self) -> None:
        """Add a new data filter."""
        user_filter = UserFilter(
            filter_type=FilterType.DATA_MATCHING,
            match_pattern=orjson.dumps({"keyword": "", "data_key": "coin"}).decode(
                "utf-8",
            ),
            action_type=ActionType.COIN_ASSOCIATION,
            action_args=orjson.dumps({"coin": "BTC"}).decode("utf-8"),
        )
        self._data_matching_layout.insertWidget(
            self._data_matching_layout.count() - 1,
            DataMatchingWidget(user_filter),
        )

    def _reset_filters(self) -> None:
        """Reset filters to match database."""
        # Delete all current filters widgets
        keyword_matching_widgets = [
            self._keyword_matching_layout.itemAt(index).widget()
            for index in range(self._keyword_matching_layout.count())
            if isinstance(
                self._keyword_matching_layout.itemAt(index).widget(),
                KeywordMatchingWidget,
            )
        ]
        for widget in keyword_matching_widgets:
            self._keyword_matching_layout.removeWidget(widget)
            widget.deleteLater()

        data_matching_widgets = [
            self._data_matching_layout.itemAt(index).widget()
            for index in range(self._data_matching_layout.count())
            if isinstance(
                self._data_matching_layout.itemAt(index).widget(),
                DataMatchingWidget,
            )
        ]
        for widget in data_matching_widgets:
            self._data_matching_layout.removeWidget(widget)
            widget.deleteLater()

        # Create new filters widget matching database
        user_filters = CONFIG.get_all_user_filters()
        for user_filter in user_filters:
            if int(user_filter.filter_type) == FilterType.KEYWORD_MATCHING:  # type: ignore
                self._keyword_matching_layout.insertWidget(
                    self._keyword_matching_layout.count() - 1,
                    KeywordMatchingWidget(user_filter),
                )
            if int(user_filter.filter_type) == FilterType.DATA_MATCHING:  # type: ignore
                self._data_matching_layout.insertWidget(
                    self._data_matching_layout.count() - 1,
                    DataMatchingWidget(user_filter),
                )

    def _save_filters(self) -> None:
        """Save filters to database."""
        for index in range(self._keyword_matching_layout.count()):
            widget = self._keyword_matching_layout.itemAt(index).widget()
            if isinstance(widget, KeywordMatchingWidget):
                widget.write_to_db()

        for index in range(self._data_matching_layout.count()):
            widget = self._data_matching_layout.itemAt(index).widget()
            if isinstance(widget, DataMatchingWidget):
                widget.write_to_db()

        Toast.show_message("Filters updated", type_=ToastType.SUCCESS)


class ColorButton(QtWidgets.QPushButton):
    """Custom Qt Widget to show a chosen color."""

    color_changed = QtCore.Signal(object)

    def __init__(self, *args: object, color: QtGui.QColor, **kwargs: dict) -> None:
        """Initialize ColorButton."""
        super().__init__(*args, **kwargs)  # type: ignore
        self.setObjectName("buttonColor")

        self._color: QtGui.QColor = color if color else QtGui.QColor(255, 0, 0)
        self.pressed.connect(self.on_color_picker)

        # Set the initial/default state.
        self.set_color(color)

    @property
    def color(self) -> QtGui.QColor:
        """Returns current color.

        Returns:
            QtGui.QColor: Color of the button.
        """
        return self._color

    def set_color(self, color: QtGui.QColor) -> None:
        """Set color."""
        if color != self._color:
            self._color = color
            self.color_changed.emit(color)

        if self._color:
            self.setStyleSheet(
                "QPushButton#buttonColor {background-color: %s;}" % self._color.name(),
            )
        else:
            self.setStyleSheet("")

    def on_color_picker(self) -> None:
        """Show color-picker dialog to select color."""
        color_dialog = QtWidgets.QColorDialog(self)
        color_dialog.setOption(
            QtWidgets.QColorDialog.ColorDialogOption.ShowAlphaChannel,
            False,
        )
        color_dialog.setOption(
            QtWidgets.QColorDialog.ColorDialogOption.DontUseNativeDialog,
            True,
        )

        if color_dialog.exec_():
            self.set_color(color_dialog.currentColor())


class BaseFilterWidget(QtWidgets.QFrame):
    """Base Widget to control filter."""

    def __init__(
        self,
        user_filter: UserFilter,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)
        self._user_filter = user_filter
        self._to_delete = True

    def on_delete(self) -> None:
        """Handle delete."""
        self.hide()
        self._to_delete = True

    def write_to_db(self) -> None:
        """Write user_filter to database."""
        if self._to_delete:
            CONFIG.delete_user_filter(self._user_filter.id)  # type: ignore
            self.deleteLater()
            return


class KeywordMatchingWidget(BaseFilterWidget):
    """Widget to control keyword matching filter."""

    def __init__(
        self,
        user_filter: UserFilter,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(user_filter=user_filter, parent=parent)

        self._sfx = QSoundEffect()

        self._main_layout = QtWidgets.QHBoxLayout(self)
        self._if_label = QtWidgets.QLabel("IF")
        self._match_pattern = QtWidgets.QLineEdit()
        self._then_label = QtWidgets.QLabel("THEN")
        self._action_combo = QtWidgets.QComboBox()
        self._sound_combo = QtWidgets.QComboBox()
        self._sound_button = QtWidgets.QPushButton()
        self._coin_line = QtWidgets.QLineEdit()
        self._color_picker = ColorButton(color=QtGui.QColor("red"))
        self._delete_btn = QtWidgets.QPushButton()

        self._setup_widgets()
        self._setup_layout()

        self.reset_to_current()
        self.setMinimumHeight(self.sizeHint().height())

    def _setup_widgets(self) -> None:
        """Configure Widgets."""
        self.setObjectName("config_item")

        self._if_label.setObjectName("title")

        self._match_pattern.setPlaceholderText("Pattern to Match...")
        self._match_pattern.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )

        for action in FILTER_ACTIONS_MAP:
            self._action_combo.addItem(action.name.capitalize(), userData=action)
        self._action_combo.currentIndexChanged.connect(self.on_action_change)

        self._then_label.setObjectName("title")

        self._coin_line.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )

        for path in list_resources_from_prefix("sfx"):
            self._sound_combo.addItem(
                path,
                userData=f":/sfx/{path}",
            )
        self._sound_button.setIcon(QtGui.QPixmap(":/icons/music"))
        self._sound_button.setProperty("class", "borderless")
        self._sound_button.setToolTip("Play Sound")
        self._sound_button.clicked.connect(self.on_play_sound)

        self._coin_line.setPlaceholderText("Coin to assign")

        self._delete_btn.setIcon(QtGui.QPixmap(":/icons/delete_icon"))
        self._delete_btn.setMinimumSize(32, 32)
        self._delete_btn.setProperty("class", "borderless")
        self._delete_btn.setToolTip("Delete Filter")
        self._delete_btn.clicked.connect(self.on_delete)

    def on_play_sound(self) -> None:
        """Play sound."""
        sfx_path = self._sound_combo.itemData(self._sound_combo.currentIndex())
        self._sfx.setSource(QtCore.QUrl.fromLocalFile(sfx_path))
        self._sfx.play()

    def on_action_change(self, index: int) -> None:
        """Handle action change."""
        action = self._action_combo.itemData(index)
        if action == ActionType.COIN_ASSOCIATION:
            self._sound_combo.setVisible(False)
            self._sound_button.setVisible(False)
            self._coin_line.setVisible(True)
            self._color_picker.setVisible(True)
        elif action == ActionType.SOUND_ASSOCIATION:
            self._sound_combo.setVisible(True)
            self._sound_button.setVisible(True)
            self._coin_line.setVisible(False)
            self._color_picker.setVisible(True)
        elif action == ActionType.IGNORE:
            self._sound_combo.setVisible(False)
            self._sound_button.setVisible(False)
            self._coin_line.setVisible(False)
            self._color_picker.setVisible(False)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addWidget(
            self._if_label,
            alignment=QtCore.Qt.AlignmentFlag.AlignHCenter,
        )
        self._main_layout.addWidget(self._match_pattern)
        self._main_layout.addWidget(
            self._then_label,
            alignment=QtCore.Qt.AlignmentFlag.AlignHCenter,
        )
        self._main_layout.addWidget(self._action_combo)
        self._main_layout.addWidget(self._sound_combo)
        self._main_layout.addWidget(self._sound_button)
        self._main_layout.addWidget(self._coin_line)
        self._main_layout.addWidget(self._color_picker)
        self._main_layout.addWidget(self._delete_btn)

        self.setLayout(self._main_layout)

    def reset_to_current(self) -> None:
        """Reset values to current filter."""
        match_pattern = orjson.loads(str(self._user_filter.match_pattern))
        if keyword := match_pattern.get("keyword", ""):
            self._match_pattern.setText(keyword)
        action_index = self._action_combo.findData(
            ActionType(int(self._user_filter.action_type)),  # type: ignore
        )
        self._action_combo.setCurrentIndex(action_index)

        action_args = orjson.loads(str(self._user_filter.action_args))

        if sound_path := action_args.get("sound_path", ""):
            sound_index = self._sound_combo.findData(sound_path)
            self._sound_combo.setCurrentIndex(sound_index)

        if coin := action_args.get("coin", ""):
            self._coin_line.setText(coin)

        if color := action_args.get("color", ""):
            self._color_picker.set_color(QtGui.QColor(*color))

        self.on_action_change(action_index)
        self._to_delete = False

    def write_to_db(self) -> None:
        """Write user_filter to database."""
        super().write_to_db()

        self._user_filter.match_pattern = orjson.dumps(  # type: ignore
            {"keyword": self._match_pattern.text()},
        ).decode("utf-8")
        self._user_filter.action_type = self._action_combo.currentData()  # type: ignore
        action_args = {}
        if self._sound_combo.isVisible():
            action_args["sound_path"] = self._sound_combo.currentData()  # type: ignore
        if self._coin_line.isVisible():
            action_args["coin"] = self._coin_line.text()
        if self._color_picker.isVisible():
            action_args["color"] = self._color_picker.color.toTuple()[0:3]  # type: ignore

        self._user_filter.action_args = orjson.dumps(action_args).decode("utf-8")  # type: ignore
        CONFIG.write_model_to_db(self._user_filter)


class DataMatchingWidget(BaseFilterWidget):
    """Widget to control data matching filter."""

    def __init__(
        self,
        user_filter: UserFilter,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(user_filter=user_filter, parent=parent)
        self._sfx = QSoundEffect()

        self._main_layout = QtWidgets.QHBoxLayout(self)
        self._if_label = QtWidgets.QLabel("IF")
        self._match_pattern = QtWidgets.QLineEdit()
        self._in_label = QtWidgets.QLabel("IN")
        self._data_field = QtWidgets.QComboBox()
        self._then_label = QtWidgets.QLabel("THEN")
        self._action_combo = QtWidgets.QComboBox()
        self._sound_combo = QtWidgets.QComboBox()
        self._sound_button = QtWidgets.QPushButton()
        self._coin_line = QtWidgets.QLineEdit()
        self._delete_btn = QtWidgets.QPushButton()

        self._setup_widgets()
        self._setup_layout()

        self.reset_to_current()
        self.setMinimumHeight(self.sizeHint().height())

    def _setup_widgets(self) -> None:
        """Configure Widgets."""
        self.setObjectName("config_item")

        self._if_label.setObjectName("title")

        self._match_pattern.setPlaceholderText("Pattern to Match...")
        self._match_pattern.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )

        self._in_label.setObjectName("title")
        valid_fields = ["title", "quoter", "coin", "source", "feed"]
        for field in valid_fields:
            self._data_field.addItem(field.capitalize(), userData=field)

        for action in FILTER_ACTIONS_MAP:
            self._action_combo.addItem(action.name.capitalize(), userData=action)
        self._action_combo.currentIndexChanged.connect(self.on_action_change)

        self._then_label.setObjectName("title")

        self._coin_line.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )

        for path in list_resources_from_prefix("sfx"):
            self._sound_combo.addItem(
                path,
                userData=f":/sfx/{path}",
            )
        self._sound_button.setIcon(QtGui.QPixmap(":/icons/music"))
        self._sound_button.setProperty("class", "borderless")
        self._sound_button.setToolTip("Play Sound")
        self._sound_button.clicked.connect(self.on_play_sound)

        self._coin_line.setPlaceholderText("Coin to assign")

        self._delete_btn.setIcon(QtGui.QPixmap(":/icons/delete_icon"))
        self._delete_btn.setMinimumSize(32, 32)
        self._delete_btn.setProperty("class", "borderless")
        self._delete_btn.setToolTip("Delete Filter")
        self._delete_btn.clicked.connect(self.on_delete)

    def on_play_sound(self) -> None:
        """Play sound."""
        sfx_path = self._sound_combo.itemData(self._sound_combo.currentIndex())
        self._sfx.setSource(QtCore.QUrl.fromLocalFile(sfx_path))
        self._sfx.play()

    def on_action_change(self, index: int) -> None:
        """Handle action change."""
        action = self._action_combo.itemData(index)
        if action == ActionType.COIN_ASSOCIATION:
            self._sound_combo.setVisible(False)
            self._sound_button.setVisible(False)
            self._coin_line.setVisible(True)
        elif action == ActionType.SOUND_ASSOCIATION:
            self._sound_combo.setVisible(True)
            self._sound_button.setVisible(True)
            self._coin_line.setVisible(False)
        elif action == ActionType.IGNORE:
            self._sound_combo.setVisible(False)
            self._sound_button.setVisible(False)
            self._coin_line.setVisible(False)

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._main_layout.addWidget(
            self._if_label,
            alignment=QtCore.Qt.AlignmentFlag.AlignHCenter,
        )
        self._main_layout.addWidget(self._match_pattern)
        self._main_layout.addWidget(
            self._in_label,
            alignment=QtCore.Qt.AlignmentFlag.AlignHCenter,
        )
        self._main_layout.addWidget(self._data_field)
        self._main_layout.addWidget(
            self._then_label,
            alignment=QtCore.Qt.AlignmentFlag.AlignHCenter,
        )
        self._main_layout.addWidget(self._action_combo)
        self._main_layout.addWidget(self._sound_combo)
        self._main_layout.addWidget(self._sound_button)
        self._main_layout.addWidget(self._coin_line)
        self._main_layout.addWidget(self._delete_btn)

        self.setLayout(self._main_layout)

    def reset_to_current(self) -> None:
        """Reset values to current filter."""
        match_pattern = orjson.loads(str(self._user_filter.match_pattern))
        if keyword := match_pattern.get("keyword", ""):
            self._match_pattern.setText(keyword)
        if data_key := match_pattern.get("data_key", ""):
            data_index = self._data_field.findData(data_key)
            self._data_field.setCurrentIndex(data_index)

        action_index = self._action_combo.findData(
            ActionType(int(self._user_filter.action_type)),  # type: ignore
        )
        self._action_combo.setCurrentIndex(action_index)

        action_args = orjson.loads(str(self._user_filter.action_args))

        if sound_path := action_args.get("sound_path", ""):
            sound_index = self._sound_combo.findData(sound_path)
            self._sound_combo.setCurrentIndex(sound_index)

        if coin := action_args.get("coin", ""):
            self._coin_line.setText(coin)

        self.on_action_change(action_index)
        self._to_delete = False

    def write_to_db(self) -> None:
        """Write user_filter to database."""
        super().write_to_db()

        self._user_filter.match_pattern = orjson.dumps(  # type: ignore
            {
                "keyword": self._match_pattern.text(),
                "data_key": self._data_field.currentData(),
            },
        ).decode("utf-8")
        self._user_filter.action_type = self._action_combo.currentData()  # type: ignore
        action_args = {}
        if self._sound_combo.isVisible():
            action_args["sound_path"] = self._sound_combo.currentData()  # type: ignore
        if self._coin_line.isVisible():
            action_args["coin"] = self._coin_line.text()

        self._user_filter.action_args = orjson.dumps(action_args).decode("utf-8")  # type: ignore
        CONFIG.write_model_to_db(self._user_filter)



"""Base fetcher for CCTX api."""

from __future__ import annotations

import asyncio
from collections import defaultdict
from decimal import Decimal
import logging
import math
from typing import TYPE_CHECKING

import pandas
from tenacity import before_sleep_log, retry, stop_after_attempt, wait_exponential

from plutus_terminal.core.exchange.base import (
    ExchangeFetcher,
)
from plutus_terminal.core.types_ import PerpsPosition, PerpsTradeDirection, PriceData
from plutus_terminal.log_utils import log_retry

if TYPE_CHECKING:
    import ccxt.pro

    from plutus_terminal.core.exchange.base import ExchangeFetcherMessageBus
    from plutus_terminal.core.types_ import PriceHistory

LOGGER = logging.getLogger(__name__)


class CCXTFetcher(ExchangeFetcher):
    """Base class for CCTX fetcher."""

    def __init__(
        self,
        message_bus: ExchangeFetcherMessageBus,
        exchange_instance: ccxt.pro.Exchange,
    ) -> None:
        """Initialize shared attributes."""
        self._cctx = exchange_instance
        self._message_bus = message_bus
        self.connection_count: dict[str, int] = defaultdict(int)
        self._watched_tickers: set[str] = set()
        self._cached_prices: dict[str, PriceData] = {}
        self._synced = True
        self.async_stop_event = asyncio.Event()

    async def fetch_price_history(
        self,
        pair: str,
        from_timestamp: int,
        resolution: str,
    ) -> PriceHistory:
        """Fetch the price history of pair for the given start and end time.

        Args:
            pair (str): Pair to fetch.
            from_timestamp (int): Timestamp for the leftmost bar in secods
            resolution (str): Bar resolution in seconds.

        Returns:
            PriceHistory: Dict with ohlcv.
        """
        # Transform seconds to miliseconds
        from_timestamp = int(from_timestamp * 1000)
        match resolution:
            case "5":
                resolution = "5m"
            case "15":
                resolution = "15m"
            case "30":
                resolution = "30m"
            case "60":
                resolution = "1h"
            case "240":
                resolution = "4h"
            case _:
                resolution = "1m"
        ohlcvs = await self._cctx.fetch_ohlcv(
            symbol=pair,
            timeframe=resolution,
            since=from_timestamp,
            limit=200,
        )
        ohlcvs = list(map(list, zip(*ohlcvs, strict=False)))
        price_history: PriceHistory = {
            "date": [pandas.Timestamp(t, unit="ms") for t in ohlcvs[0]],
            "open": ohlcvs[1],
            "high": ohlcvs[2],
            "low": ohlcvs[3],
            "close": ohlcvs[4],
            "volume": ohlcvs[5],  # type: ignore
        }
        LOGGER.debug(
            "Fetched price history for %s from %s",
            pair,
            from_timestamp,
        )
        return price_history

    async def subscribe_to_price(self, pair: str, force: bool = False) -> None:
        """Subscribe to receive price updates for the given pair.

        Args:
            pair (str): Pair name e.g BTC/USDT.
            force (bool, optional): Force subscribe. Defaults to False.
        """
        self.connection_count[pair] += 1
        if self.connection_count[pair] > 1 and not force:
            return
        self._watched_tickers.add(pair)
        LOGGER.info("Subscribed to receive price updates of: %s", pair)

    async def resubscribe_on_going_connections(self) -> None:
        """Resubscribe on going connections."""
        LOGGER.warning("Re-subscribing to on going connection...")
        for pair, count in self.connection_count.items():
            if count >= 1:
                await self.subscribe_to_price(pair, force=True)
                # Reduce 1 since this connection should already be present
                self.connection_count[pair] -= 1

    async def unsubscribe_to_price(self, pair: str, force: bool = False) -> None:
        """Unsubscribe to receive price updates for the given pair.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
            force (bool, optional): Force unsubscribe. Defaults to False.
        """
        self.connection_count[pair] -= 1
        if self.connection_count[pair] > 0 and not force:
            return
        self._watched_tickers.discard(pair)
        del self._cached_prices[pair]
        LOGGER.info("Unsubscribed to receive price updates of: %s", pair)

    async def receive_subscribed_prices(self) -> None:
        """Receive live data of subscribed prices from exchange."""
        while not self.async_stop_event.is_set():
            try:
                tickers = await self._cctx.watch_tickers(list(self._watched_tickers))
                for ticker in tickers.values():  # type: ignore
                    pair = ticker["symbol"]
                    self._cached_prices[pair] = PriceData(
                        {
                            "price": Decimal(ticker["last"]),
                            "date": pandas.Timestamp(ticker["timestamp"], unit="ms"),
                            "volume": ticker["quoteVolume"],
                        },
                    )
                    self._message_bus.subscribed_prices_signal.emit(self._cached_prices)
            except asyncio.CancelledError:
                LOGGER.debug("Watch all positions cancelled.")
                raise
            except Exception:
                LOGGER.exception("Unexpected error while receiving prices")

    @retry(
        reraise=True,
        stop=stop_after_attempt(15),
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_price_at_time(self, pair: str, timestamp: int) -> PriceData:
        """Fetch price at current time.

        Args:
            pair (str): Pair name e.g Crypto.BTC/USD.
            timestamp (int): Unix timestamp in seconds.

        Returns:
            PriceData: Price data for fetched pair.
        """
        LOGGER.debug("Fetching %s price at time: %s", pair, timestamp)
        rounded_timestamp = math.floor(timestamp / 60) * 60
        timestamp = int(rounded_timestamp * 1000)
        ticker = await self._cctx.fetch_ohlcv(
            symbol=pair,
            timeframe="1m",
            since=timestamp,
            limit=1,
        )
        return PriceData(
            {
                "price": Decimal(ticker[0][4]),
                "date": pandas.Timestamp(ticker[0][0], unit="ms"),
                "volume": ticker[0][6],
            },
        )

    async def watch_all_positions(self) -> None:
        """Watch all open positions for all available pairs."""
        while not self.async_stop_event.is_set():
            all_positions = []
            try:
                all_positions.extend(await self._cctx.fetch_positions())
            except asyncio.CancelledError:
                LOGGER.debug("Watch all positions cancelled.")
                raise
            except Exception:
                LOGGER.exception("Unexpected error while watching positions")
            self._message_bus.positions_signal.emit(all_positions)

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=5),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_all_positions(self) -> list[PerpsPosition]:
        """Fetch all open positions for all available pairs."""
        all_positions: list[PerpsPosition] = []
        positions = await self._cctx.fetch_positions()
        for position in positions:  # type: ignore
            position_size_stable = Decimal(position["entryPrice"]) * Decimal(
                position["contracts"] / position["leverage"],
            )
            collateral = position_size_stable / Decimal(position["leverage"])
            persp_position = PerpsPosition(
                {
                    "pair": position["symbol"],
                    "id": position["id"],
                    "position_size_stable": position_size_stable,
                    "collateral_stable": collateral,
                    "open_price": Decimal(position["entryPrice"]),
                    "trade_direction": PerpsTradeDirection[position["side"].upper()],
                    "leverage": Decimal(position["leverage"]),
                },
            )
            all_positions.append(persp_position)
        return all_positions

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def fetch_current_price(self, pair: str) -> PriceData:
        """Fetch current price of given pair."""
        ticker = await self._cctx.fetch_ticker(symbol=pair)
        return PriceData(
            {
                "price": Decimal(ticker["last"]),
                "date": pandas.Timestamp(ticker["timestamp"], unit="ms"),
                "volume": ticker["quoteVolume"],
            },
        )

    async def stop_async(self) -> None:
        """Stop infinite loops and close connections."""
        self.async_stop_event.set()



"""Dict of valid exchanges."""

from plutus_terminal.core.exchange.bitget.exchange import BitgetExchange
from plutus_terminal.core.exchange.foxify.exchange import FoxifyExchange
from plutus_terminal.core.exchange.foxify.funded_exchange import FoxifyFundedExchange

VALID_EXCHANGES = {"foxify": FoxifyExchange, "foxify_funded": FoxifyFundedExchange}



"""Foxify Funded Exchange trader."""

from eth_account.signers.local import LocalAccount
from qasync import logging
from tenacity import before_sleep_log, retry, wait_exponential
from web3.types import Gwei

from plutus_terminal.core.exchange.foxify import utils as foxify_utils
from plutus_terminal.core.exchange.foxify.trader import FoxifyTrader
from plutus_terminal.log_utils import log_retry

LOGGER = logging.getLogger(__name__)


class FoxifyFundedTrader(FoxifyTrader):
    """Foxify Funded Exchange trader."""

    def __init__(
        self,
        pair_map: dict[str, str],
        web3_account: LocalAccount,
        extra_gas: Gwei,
    ) -> None:
        """Initialize shared attributes."""
        super().__init__(pair_map, web3_account, extra_gas)

    @retry(
        wait=wait_exponential(multiplier=1, min=0.15, max=1),
        before_sleep=before_sleep_log(LOGGER, logging.DEBUG),
        retry_error_callback=log_retry(LOGGER),
    )
    async def init_async(self) -> None:
        """Init async shared attributes."""
        await super().init_async()

        funded_factory_contract = foxify_utils.build_funded_factory_contract(
            self.web3_provider,
        )

        self._funded_trader_address = self.web3_provider.to_checksum_address(
            await funded_factory_contract.functions.traderContracts(
                self.web3_account.address,
            ).call(),
        )

        self._funded_trader_contract = foxify_utils.build_funded_trader_contract(
            self._funded_trader_address,
            self.web3_provider,
        )

        # Override position router and order book to point to trader contract
        self._position_router_contract = self._funded_trader_contract
        self._order_book_contract = self._funded_trader_contract

        # override receiver address
        self._receiver_address = self._funded_trader_address



"""Widget to buy options with strategy."""

from __future__ import annotations

import asyncio
from decimal import Decimal
from functools import partial
from typing import TYPE_CHECKING, Optional

from PySide6 import QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtGui import QPixmap
from qasync import asyncSlot

from plutus_terminal.core.config import CONFIG
from plutus_terminal.core.types_ import (
    OptionsDirection,
    OptionsDuration,
    OptionsRisk,
)
from plutus_terminal.ui.widgets.options_table import OptionsTableModel, OptionsTableView

if TYPE_CHECKING:
    from plutus_terminal.core.exchange.base import ExchangeBase


class OptionsWidget(QtWidgets.QWidget):
    """Widget to buy options with strategy."""

    def __init__(
        self,
        exchange: ExchangeBase,
        parent: Optional[QtWidgets.QWidget] = None,
    ) -> None:
        """Initialize widget."""
        super().__init__(parent=parent)
        self._exchange = exchange
        self._stop_update = asyncio.Event()
        self._update_task: Optional[asyncio.Task] = None

        self.main_layout = QtWidgets.QGridLayout(self)

        self._top_bar_layout = QtWidgets.QHBoxLayout()
        self._top_bar_icon = QtWidgets.QLabel()
        self._top_bar_title = QtWidgets.QLabel("Options Strategy")
        self._top_bar_preview = QtWidgets.QRadioButton()
        self._top_bar_line = QtWidgets.QFrame()

        self.token_label = QtWidgets.QLabel("Token:")
        self.token_combo = QtWidgets.QComboBox()
        self.amount_label = QtWidgets.QLabel("Amount to buy:")
        self.amount_spin = QtWidgets.QDoubleSpinBox()
        self.rate_min_label = QtWidgets.QLabel("Rate Min:")
        self.rate_min_spin = QtWidgets.QDoubleSpinBox()
        self.available_min_label = QtWidgets.QLabel("Available Min:")
        self.available_min_spin = QtWidgets.QDoubleSpinBox()
        self.percent_min_label = QtWidgets.QLabel("Percent Min:")
        self.percent_min_group = QtWidgets.QButtonGroup()
        self.percent_min_layout = QtWidgets.QHBoxLayout()
        self.percent_max_label = QtWidgets.QLabel("Percent Max:")
        self.percent_max_group = QtWidgets.QButtonGroup()
        self.percent_max_layout = QtWidgets.QHBoxLayout()
        self.duration_min_label = QtWidgets.QLabel("Duration Min:")
        self.duration_min_group = QtWidgets.QButtonGroup()
        self.duration_min_layout = QtWidgets.QHBoxLayout()
        self.duration_max_label = QtWidgets.QLabel("Duration Max:")
        self.duration_max_group = QtWidgets.QButtonGroup()
        self.duration_max_layout = QtWidgets.QHBoxLayout()
        self.risk_label = QtWidgets.QLabel("Risk:")
        self.risk_group = QtWidgets.QButtonGroup()
        self.risk_layout = QtWidgets.QHBoxLayout()

        self.buy_up_button = QtWidgets.QPushButton("Buy Up")
        self.buy_down_button = QtWidgets.QPushButton("Buy Down")

        self._tab_table = QtWidgets.QTabWidget()
        self._long_table_model = OptionsTableModel()
        self._long_table_view = OptionsTableView()
        self._long_posible_buy = QtWidgets.QLabel()
        self._short_table_model = OptionsTableModel()
        self._short_table_view = OptionsTableView()
        self._short_posible_buy = QtWidgets.QLabel()

        self._percent_text = ["0%", "0.10%", "0.25%", "0.50%", "1%", "3%", "5%"]

        self._spin_config_map = {
            self.amount_spin: "options_amount",
            self.rate_min_spin: "options_rate_min",
            self.available_min_spin: "options_available_min",
        }

        self._button_grp_config_map = {
            self.percent_min_group: "options_percent_min",
            self.percent_max_group: "options_percent_max",
            self.duration_min_group: "options_duration_min",
            self.duration_max_group: "options_duration_max",
            self.risk_group: "options_risk",
        }

        self._setup_widgets()
        self._setup_layout()
        self._setup_for_has_options()

    def _setup_widgets(self) -> None:  # noqa: C901, PLR0915
        """Create widgets."""
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred,
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
        )

        self._top_bar_icon.setPixmap(QPixmap(":icons/options_icon"))
        self._top_bar_icon.setMinimumHeight(35)
        self._top_bar_title.setObjectName("title")
        self._top_bar_title.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed,
            QtWidgets.QSizePolicy.Policy.Fixed,
        )
        self._top_bar_title.setAlignment(
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        )

        self._top_bar_preview.setIcon(QPixmap(":/icons/preview_icon"))
        self._top_bar_preview.setChecked(
            CONFIG.get_gui_settings("options_show_preview"),
        )
        self._top_bar_preview.setToolTip(
            "Show options that will be bought with current strategy.",
        )
        self._top_bar_preview.toggled.connect(self.show_preview_toggled)

        self._top_bar_line.setObjectName("divisor")
        self._top_bar_line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self._top_bar_line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)

        self.amount_spin.setMaximum(1_000_000_000)
        self.amount_spin.setMinimum(1)
        self.amount_spin.setValue(CONFIG.options_amount)
        self.amount_spin.valueChanged.connect(
            partial(setattr, CONFIG, "options_amount"),
        )

        self.rate_min_spin.setMaximum(100)
        self.rate_min_spin.setMinimum(0.01)
        self.rate_min_spin.setSingleStep(0.01)
        self.rate_min_spin.setValue(CONFIG.options_rate_min)
        self.rate_min_spin.valueChanged.connect(
            partial(setattr, CONFIG, "options_rate_min"),
        )

        self.available_min_spin.setMaximum(100_000_000_000)
        self.available_min_spin.setMinimum(2.5)
        self.available_min_spin.setValue(CONFIG.options_available_min)
        self.available_min_spin.valueChanged.connect(
            partial(setattr, CONFIG, "options_available_min"),
        )

        for percent in self._percent_text:
            button = QtWidgets.QRadioButton(percent)
            if percent == CONFIG.options_percent_min:
                button.setChecked(True)
            self.percent_min_group.addButton(button)
            self.percent_min_layout.addWidget(button)

        self.percent_min_group.buttonClicked.connect(
            partial(self._set_config_property, "options_percent_min"),
        )

        for percent in self._percent_text:
            button = QtWidgets.QRadioButton(percent)
            if percent == CONFIG.options_percent_max:
                button.setChecked(True)
            self.percent_max_group.addButton(button)
            self.percent_max_layout.addWidget(button)

        self.percent_max_group.buttonClicked.connect(
            partial(self._set_config_property, "options_percent_max"),
        )

        for risk in OptionsDuration:
            button = QtWidgets.QRadioButton(risk.name)
            if risk == CONFIG.options_duration_min:
                button.setChecked(True)
            self.duration_min_group.addButton(button)
            self.duration_min_layout.addWidget(button)

        self.duration_min_group.buttonClicked.connect(
            partial(
                self._set_config_property_enum,
                "options_duration_min",
                OptionsDuration,
            ),
        )

        for duration in OptionsDuration:
            button_duration = QtWidgets.QRadioButton(duration.name)
            if duration == CONFIG.options_duration_max:
                button_duration.setChecked(True)
            self.duration_max_group.addButton(button_duration)
            self.duration_max_layout.addWidget(button_duration)

        self.duration_max_group.buttonClicked.connect(
            partial(
                self._set_config_property_enum,
                "options_duration_max",
                OptionsDuration,
            ),
        )

        for risk in OptionsRisk:
            button_risk = QtWidgets.QRadioButton(risk.name)
            if risk.name == CONFIG.options_risk.name:
                button_risk.setChecked(True)
            self.risk_group.addButton(button_risk)
            self.risk_layout.addWidget(button_risk)

        self.risk_group.buttonClicked.connect(
            partial(self._set_config_property_enum, "options_risk", OptionsRisk),
        )

        self.buy_up_button.setProperty("class", "LONG")
        self.buy_up_button.setMinimumHeight(35)
        self.buy_up_button.clicked.connect(
            partial(self._buy_options, OptionsDirection.UP),
        )

        self.buy_down_button.setProperty("class", "SHORT")
        self.buy_down_button.setMinimumHeight(35)
        self.buy_down_button.clicked.connect(
            partial(self._buy_options, OptionsDirection.DOWN),
        )

        self._long_posible_buy.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._short_posible_buy.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self._tab_table.setVisible(CONFIG.get_gui_settings("options_show_preview"))

        self._long_table_view.setModel(self._long_table_model)
        self._short_table_view.setModel(self._short_table_model)

    def _setup_for_has_options(self) -> None:
        """Configure widget for has options enabled."""
        if self._exchange.has_options():
            avilable_pairs = self._exchange.options.fetch_available_pairs()
            self.buy_up_button.setEnabled(True)
            self.buy_down_button.setEnabled(True)
            self._top_bar_preview.setEnabled(True)
        else:
            avilable_pairs = []
            self.buy_up_button.setEnabled(False)
            self.buy_down_button.setEnabled(False)
            self._top_bar_preview.setEnabled(False)
        self.token_combo.addItems(avilable_pairs)

        # Start update task to show preivew if enabled
        if self._exchange.has_options() and CONFIG.get_gui_settings(
            "options_show_preview",
        ):  # type: ignore
            self._update_task = asyncio.create_task(self.update_previews())

    @asyncSlot()
    async def _set_config_property(
        self,
        property_name: str,
        button: QtWidgets.QRadioButton,
    ) -> None:
        """Set property on config."""
        setattr(CONFIG, property_name, button.text())
        await self._update_previews()

    @asyncSlot()
    async def _set_config_property_enum(
        self,
        property_name: str,
        enum: type[OptionsRisk | OptionsDuration],
        button: QtWidgets.QRadioButton,
    ) -> None:
        """Set property on config."""
        setattr(CONFIG, property_name, enum[button.text()].value)
        await self._update_previews()

    @asyncSlot()
    async def _buy_options(self, direction: OptionsDirection) -> None:
        """Buy options."""
        await self._exchange.buy_options_with_strategy(
            direction,
            Decimal(self.amount_spin.value()),
            self.token_combo.currentText(),
        )

    def _setup_layout(self) -> None:
        """Configure layout."""
        self._top_bar_layout.addWidget(self._top_bar_icon)
        self._top_bar_layout.addWidget(self._top_bar_title)
        self._top_bar_layout.addStretch()
        self._top_bar_layout.addWidget(self._top_bar_preview)
        self.main_layout.addLayout(self._top_bar_layout, 0, 0, 1, 2)
        self.main_layout.addWidget(self._top_bar_line, 1, 0, 1, 2)
        self.main_layout.addWidget(self.token_label, 2, 0)
        self.main_layout.addWidget(self.token_combo, 2, 1)
        self.main_layout.addWidget(self.amount_label, 3, 0)
        self.main_layout.addWidget(self.amount_spin, 3, 1)
        self.main_layout.addWidget(self.rate_min_label, 4, 0)
        self.main_layout.addWidget(self.rate_min_spin, 4, 1)
        self.main_layout.addWidget(self.available_min_label, 5, 0)
        self.main_layout.addWidget(self.available_min_spin, 5, 1)
        self.main_layout.addWidget(self.percent_min_label, 6, 0, 1, 2)
        self.main_layout.addLayout(self.percent_min_layout, 7, 0, 1, 2)
        self.main_layout.addWidget(self.percent_max_label, 8, 0, 1, 2)
        self.main_layout.addLayout(self.percent_max_layout, 9, 0, 1, 2)
        self.main_layout.addWidget(self.duration_min_label, 10, 0, 1, 2)
        self.main_layout.addLayout(self.duration_min_layout, 11, 0, 1, 2)
        self.main_layout.addWidget(self.duration_max_label, 12, 0, 1, 2)
        self.main_layout.addLayout(self.duration_max_layout, 13, 0, 1, 2)
        self.main_layout.addWidget(self.risk_label, 14, 0, 1, 2)
        self.main_layout.addLayout(self.risk_layout, 15, 0, 1, 2)
        self.main_layout.addWidget(self.buy_up_button, 16, 0)
        self.main_layout.addWidget(self.buy_down_button, 16, 1)

        long_tab_layout = QtWidgets.QVBoxLayout()
        long_tab_layout.addWidget(self._long_table_view)
        long_tab_layout.addWidget(self._long_posible_buy)
        long_tab_widget = QtWidgets.QWidget()
        long_tab_widget.setContentsMargins(0, 0, 0, 0)
        long_tab_widget.setLayout(long_tab_layout)
        self._tab_table.addTab(long_tab_widget, "Buy Up Preview")

        short_tab_layout = QtWidgets.QVBoxLayout()
        short_tab_layout.addWidget(self._short_table_view)
        short_tab_layout.addWidget(self._short_posible_buy)
        short_tab_widget = QtWidgets.QWidget()
        short_tab_widget.setContentsMargins(0, 0, 0, 0)
        short_tab_widget.setLayout(short_tab_layout)
        self._tab_table.addTab(short_tab_widget, "Buy Down Preview")

        self.main_layout.addWidget(self._tab_table, 17, 0, 1, 2)
        self.main_layout.setRowStretch(18, 1)

    @asyncSlot()
    async def show_preview_toggled(self, checked: bool) -> None:
        """Show preview."""
        CONFIG.set_gui_settings("options_show_preview", checked)
        self._tab_table.setVisible(checked)
        if checked:
            if self._update_task is not None:
                await self._update_task
            self._stop_update.clear()
            self._update_task = asyncio.create_task(self.update_previews())
        else:
            self._stop_update.set()
            if self._update_task is not None:
                await self._update_task

    async def update_previews(self) -> None:
        """Infinite async loop to update options previews each 5 seconds."""
        while not self._stop_update.is_set():
            await self._update_previews()
            await asyncio.sleep(5)

    async def _update_previews(self) -> None:
        """Update strategy previews based on current values."""
        up_data = await self._exchange.options.filter_with_strategy(
            OptionsDirection.UP,
            Decimal(self.amount_spin.value()),
            self.token_combo.currentText(),
        )
        self._long_table_model.update_data(up_data)
        if up_data.empty:
            total_available = 0.0
        else:
            total_available = up_data["available"].sum() / 10**6
            if total_available > self.amount_spin.value():
                total_available = self.amount_spin.value()
        self._long_posible_buy.setText(
            f"<u><b>${total_available:,.2f}</b> "
            "to be bought in "
            f"<b>{up_data.shape[0]}</b> different options</u>",
        )
        down_data = await self._exchange.options.filter_with_strategy(
            OptionsDirection.DOWN,
            Decimal(self.amount_spin.value()),
            self.token_combo.currentText(),
        )
        self._short_table_model.update_data(down_data)
        if down_data.empty:
            total_available = 0
        else:
            total_available = down_data["available"].sum() / 10**6
            if total_available > self.amount_spin.value():
                total_available = self.amount_spin.value()
        self._short_posible_buy.setText(
            f"<u><b>${total_available:,.2f}</b> "
            "to be bought in "
            f"<b>{down_data.shape[0]}</b> different options</u>",
        )

    def on_new_exchange(self, new_exchange: ExchangeBase) -> None:
        """Update info based on new exchange.

        Args:
            new_exchange (ExchangeBase): New exchangeBase.
        """
        self._stop_update.set()
        self._exchange = new_exchange
        self._setup_for_has_options()

    def on_new_account(self) -> None:
        """Update widgets  on new account."""
        # Update spin box values
        for spin, attr in self._spin_config_map.items():
            spin.blockSignals(True)
            spin.setValue(getattr(CONFIG, attr))
            spin.blockSignals(False)

        for button_grp, attr in self._button_grp_config_map.items():
            button_grp.blockSignals(True)
            for button in button_grp.buttons():
                if button.text() == getattr(CONFIG, attr):
                    button.setChecked(True)
            button_grp.blockSignals(False)
